"""Integration test data_structures_entry"""

import pytest

from pysonos.data_structures_entry import from_didl_string
from pysonos.data_structures import (
    DidlMusicTrack,
    DidlMusicArtist,
    DidlMusicGenre,
    DidlContainer,
    DidlPlaylistContainer,
    DidlMusicAlbum,
    DidlPerson,
    DidlAudioBroadcast,
)


from conftest import DataLoader

DATA_LOADER = DataLoader("data_structures_entry_integration")


TEST_ITEMS_DATA = (
    (DidlMusicTrack, *DATA_LOADER.load_xml_and_json("track")),
    (DidlMusicAlbum, *DATA_LOADER.load_xml_and_json("album")),
    (DidlMusicArtist, *DATA_LOADER.load_xml_and_json("artist")),
    (DidlMusicGenre, *DATA_LOADER.load_xml_and_json("genre")),
    (DidlContainer, *DATA_LOADER.load_xml_and_json("share")),
    (DidlPlaylistContainer, *DATA_LOADER.load_xml_and_json("playlist")),
    (DidlAudioBroadcast, *DATA_LOADER.load_xml_and_json("audio_broadcast")),
)


@pytest.mark.parametrize(
    "klass,didl_xml_string,data",
    TEST_ITEMS_DATA,
    ids=[i[0].__name__ for i in TEST_ITEMS_DATA],
)
def test_items(klass, didl_xml_string, data):
    """Test standard DIDL item loading"""
    item = from_didl_string(didl_xml_string)[0]
    assert item.__class__ is klass
    for key, value in data.items():
        assert getattr(item, key) == value


# FIXME For the tests below that doesn't have a dedicated JSON data file, the
# XML used is faked, created by just editing the didl_class in a copy of the
# XML of the class it is derived from and therefore also compared to the same
# data as the class it is derived from. It should be replaced with data
# captured in the wild.
KNOWN_VENDOR_EXTENDED_CLASSES = (
    (
        "DidlRecentShow",
        DidlMusicTrack,
        DATA_LOADER.load_xml("recent_show.xml"),
        DATA_LOADER.load_json("track.json"),
    ),
    (
        "DidlMusicAlbumFavorite",
        DidlMusicAlbum,
        DATA_LOADER.load_xml("album_favorite.xml"),
        DATA_LOADER.load_json("album.json"),
    ),
    (
        "DidlMusicAlbumCompilation",
        DidlMusicAlbum,
        DATA_LOADER.load_xml("album_compilation.xml"),
        DATA_LOADER.load_json("album.json"),
    ),
    (
        "DidlComposer",
        DidlPerson,
        DATA_LOADER.load_xml("composer.xml"),
        DATA_LOADER.load_json("composer.json"),
    ),
    (
        "DidlAlbumList",
        DidlContainer,
        DATA_LOADER.load_xml("album_list.xml"),
        DATA_LOADER.load_json("share.json"),
    ),
    (
        "DidlSameArtist",
        DidlPlaylistContainer,
        DATA_LOADER.load_xml("same_artist.xml"),
        DATA_LOADER.load_json("playlist.json"),
    ),
    (
        "DidlPlaylistContainerFavorite",
        DidlPlaylistContainer,
        DATA_LOADER.load_xml("playlist_favorite.xml"),
        DATA_LOADER.load_json("playlist.json"),
    ),
    (
        "DidlPlaylistContainerTracklist",
        DidlPlaylistContainer,
        DATA_LOADER.load_xml("tracklist.xml"),
        DATA_LOADER.load_json("playlist.json"),
    ),
    (
        "DidlRadioShow",
        DidlContainer,
        DATA_LOADER.load_xml("radio_show.xml"),
        DATA_LOADER.load_json("radio_show.json"),
    ),
)

# There doesn't exist any tests of this kind yet for:
# "object.item.audioItem.audioBroadcast.sonos-favorite"
# "DidlAudioBroadcastFavorite",
#
# "object.itemobject.item.sonos-favorite"
# "DidlFavorite",


@pytest.mark.parametrize(
    "class_name,base_class,didl_xml_string,data",
    KNOWN_VENDOR_EXTENDED_CLASSES,
    ids=[i[0] for i in KNOWN_VENDOR_EXTENDED_CLASSES],
)
def test_vendor_extended_didl_class(class_name, base_class, didl_xml_string, data):
    """Test that vendor extended DIDL classes have proper names and base class"""
    item = from_didl_string(didl_xml_string)[0]

    # Test inheritance
    assert item.__class__.__name__ == class_name
    assert base_class is item.__class__.__bases__[0]
    assert base_class._translation == item._translation



"""Classes to handle Sonos UPnP Events and Subscriptions using asyncio.

The `Subscription` class from this module will be used in
:py:mod:`soco.services` if `config.EVENTS_MODULE` is set
to point to this module.

Example:

    Run this code, and change your volume, tracks etc::

        import logging

        logging.basicConfig()
        import soco
        import asyncio
        from pprint import pprint

        from soco import events_asyncio

        soco.config.EVENTS_MODULE = events_asyncio


        def print_event(event):
            try:
                pprint(event.variables)
            except Exception as e:
                print("There was an error in print_event:", e)


        def _get_device():
            device = soco.discover().pop().group.coordinator
            print(device.player_name)
            return device


        async def main():
            # pick a device at random and use it to get
            # the group coordinator
            loop = asyncio.get_event_loop()
            device = await loop.run_in_executor(None, _get_device)
            sub = await device.renderingControl.subscribe()
            sub2 = await device.avTransport.subscribe()
            sub.callback = print_event
            sub2.callback = print_event

            async def before_shutdown():
                await sub.unsubscribe()
                await sub2.unsubscribe()
                await events_asyncio.event_listener.async_stop()

            await asyncio.sleep(1)
            print("Renewing subscription..")
            await sub.renew()

            await asyncio.sleep(100)
            await before_shutdown()


        if __name__ == "__main__":
            asyncio.run(main())

"""

import logging
import socket
import sys
import time
import asyncio

try:
    from aiohttp import ClientSession, web
except ImportError as error:
    print(
        """ImportError: {}:
    Use of the SoCo events_asyncio module requires the 'aiohttp'
    package and its dependencies to be installed. aiohttp is not
    installed with SoCo by default due to potential issues installing
    the dependencies 'mutlidict' and 'yarl' on some platforms.
    See: https://github.com/SoCo/SoCo/issues/819""".format(
            error
        )
    )
    sys.exit(1)

# Event is imported for compatibility with events.py
# pylint: disable=unused-import
from .events_base import Event  # noqa: F401

from .events_base import (  # noqa: E402
    get_listen_ip,
    parse_event_xml,
    EventNotifyHandlerBase,
    EventListenerBase,
    SubscriptionBase,
    SubscriptionsMap,
)

from .exceptions import SoCoException  # noqa: E402

log = logging.getLogger(__name__)  # pylint: disable=C0103


class EventNotifyHandler(EventNotifyHandlerBase):
    """Handles HTTP ``NOTIFY`` Verbs sent to the listener server.
    Inherits from `soco.events_base.EventNotifyHandlerBase`.
    """

    def __init__(self):
        super().__init__()
        # The SubscriptionsMapAio instance created when this module is
        # imported. This is referenced by
        # soco.events_base.EventNotifyHandlerBase.
        self.subscriptions_map = subscriptions_map

    async def notify(self, request):
        """Serve a ``NOTIFY`` request by calling `handle_notification`
        with the headers and content.
        """
        content = await request.text()
        seq = request.headers["seq"]  # Event sequence number
        sid = request.headers["sid"]  # Event Subscription Identifier
        # find the relevant service from the sid
        # pylint: disable=no-member
        subscription = self.subscriptions_map.get_subscription(sid)
        # It might have been removed by another thread
        if subscription:
            timestamp = time.time()
            service = subscription.service
            self.log_event(seq, service.service_id, timestamp)
            log.debug("Event content: %s", content)
            if "x-sonos-http" in content:
                # parse_event_xml will generate I/O if
                # x-sonos-http is in the content
                variables = await asyncio.get_event_loop().run_in_executor(
                    None, parse_event_xml, content
                )
            else:
                variables = parse_event_xml(content)

            # Build the Event object
            event = Event(sid, seq, service, timestamp, variables)
            # pass the event details on to the service so it can update
            # its cache.
            # pylint: disable=protected-access
            service._update_cache_on_event(event)
            # Pass the event on for handling
            subscription.send_event(event)
        else:
            log.debug("No service registered for %s", sid)

        return web.Response(text="OK", status=200)

    # pylint: disable=no-self-use, missing-docstring
    def log_event(self, seq, service_id, timestamp):
        log.debug("Event %s received for %s service at %s", seq, service_id, timestamp)


class EventListener(EventListenerBase):  # pylint: disable=too-many-instance-attributes
    """The Event Listener.

    Runs an http server which is an endpoint for ``NOTIFY``
    requests from Sonos devices. Inherits from
    `soco.events_base.EventListenerBase`.
    """

    def __init__(self):
        super().__init__()
        self.sock = None
        self.ip_address = None
        self.port = None
        self.runner = None
        self.site = None
        self.session = None
        self.start_lock = None

    def start(self, any_zone):
        """A stub since the first subscribe calls async_start."""
        return

    def listen(self, ip_address):
        """A stub since since async_listen is used."""
        return

    async def async_start(self, any_zone):
        """Start the event listener listening on the local machine under the lock.

        Args:
            any_zone (SoCo): Any Sonos device on the network. It does not
                matter which device. It is used only to find a local IP
                address reachable by the Sonos net.

        """
        if not self.start_lock:
            self.start_lock = asyncio.Lock()
        async with self.start_lock:
            if self.is_running:
                return
            # Use configured IP address if there is one, else detect
            # automatically.
            ip_address = get_listen_ip(any_zone.ip_address)
            if not ip_address:
                log.exception("Could not start Event Listener: check network.")
                # Otherwise, no point trying to start server
                return
            port = await self.async_listen(ip_address)
            if not port:
                return
            self.address = (ip_address, port)
            self.session = ClientSession(raise_for_status=True)
            self.is_running = True
            log.debug("Event Listener started")

    async def async_listen(self, ip_address):
        """Start the event listener listening on the local machine at
        port 1400 (default). If this port is unavailable, the
        listener will attempt to listen on the next available port,
        within a range of 100.

        Make sure that your firewall allows connections to this port.

        This method is called by `soco.events_base.EventListenerBase.start`

        Handling of requests is delegated to an instance of the
        `EventNotifyHandler` class.

        Args:
            ip_address (str): The local network interface on which the server
                should start listening.
        Returns:
            int: The port on which the server is listening.

        Note:
            The port on which the event listener listens is configurable.
            See `config.EVENT_LISTENER_PORT`
        """
        for port_number in range(
            self.requested_port_number, self.requested_port_number + 100
        ):
            try:
                if port_number > self.requested_port_number:
                    log.debug("Trying next port (%d)", port_number)
                # pylint: disable=no-member
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.bind((ip_address, port_number))
                sock.listen(200)
                self.sock = sock
                self.port = port_number
                break
            # pylint: disable=invalid-name
            except socket.error as e:
                log.warning("Could not bind to %s:%s: %s", ip_address, port_number, e)
                continue

        if not self.port:
            return None

        self.ip_address = ip_address
        await self._async_start()
        return self.port

    async def _async_start(self):
        """Start the site."""
        handler = EventNotifyHandler()
        app = web.Application()
        app.add_routes([web.route("notify", "", handler.notify)])
        self.runner = web.AppRunner(app)
        await self.runner.setup()
        self.site = web.SockSite(self.runner, self.sock)
        await self.site.start()
        log.debug("Event listener running on %s", (self.ip_address, self.port))

    async def async_stop(self):
        """Stop the listener."""
        if self.site:
            await self.site.stop()
            self.site = None
        if self.runner:
            await self.runner.cleanup()
            self.runner = None
        if self.session:
            await self.session.close()
            self.session = None
        if self.sock:
            self.sock.close()
            self.sock = None
        self.port = None
        self.ip_address = None

    # pylint: disable=unused-argument
    def stop_listening(self, address):
        """Stop the listener."""
        asyncio.ensure_future(self.async_stop())


class Subscription(SubscriptionBase):
    """A class representing the subscription to a UPnP event.
    Inherits from `soco.events_base.SubscriptionBase`.
    """

    def __init__(self, service, callback=None):
        """
        Args:
            service (Service): The SoCo `Service` to which the subscription
                 should be made.
            event_queue (:class:`~queue.Queue`): A queue on which received
                events will be put. If not specified, a queue will be
                created and used.

        """
        super().__init__(service, None)
        #: :py:obj:`function`: callback function to be called whenever an
        #: `Event` is received. If it is set and is callable, the callback
        #: function will be called with the `Event` as the only parameter and
        #: the Subscription's event queue won't be used.
        self.callback = callback
        # The SubscriptionsMapAio instance created when this module is
        # imported. This is referenced by soco.events_base.SubscriptionBase.
        self.subscriptions_map = subscriptions_map
        # The EventListener instance created when this module is imported.
        # This is referenced by soco.events_base.SubscriptionBase.
        self.event_listener = event_listener
        # Used to keep track of the auto_renew loop
        self._auto_renew_task = None

    # pylint: disable=arguments-differ
    def subscribe(self, requested_timeout=None, auto_renew=False, strict=True):
        """Subscribe to the service.

        If requested_timeout is provided, a subscription valid for that number
        of seconds will be requested, but not guaranteed. Check
        `timeout` on return to find out what period of validity is
        actually allocated.

        This method calls `events_base.SubscriptionBase.subscribe`.

        Note:
            SoCo will try to unsubscribe any subscriptions which are still
            subscribed on program termination, but it is good practice for
            you to clean up by making sure that you call :meth:`unsubscribe`
            yourself.

        Args:
            requested_timeout(int, optional): The timeout to be requested.
            auto_renew (bool, optional): If `True`, renew the subscription
                automatically shortly before timeout. Default `False`.
            strict (bool, optional): If True and an Exception occurs during
                execution, the Exception will be raised or, if False, the
                Exception will be logged and the Subscription instance will be
                returned. Default `True`.

        Returns:
            `Subscription`: The Subscription instance.

        """
        self.subscriptions_map.subscribing()
        future = asyncio.Future()
        subscribe = super().subscribe

        async def _async_wrap_subscribe():
            try:
                if not self.event_listener.is_running:
                    await self.event_listener.async_start(self.service.soco)
                await subscribe(requested_timeout, auto_renew)
                future.set_result(self)
            except SoCoException as ex:
                future.set_exception(ex)
            except Exception as exc:  # pylint: disable=broad-except
                self._cancel_subscription(exc)
                if strict:
                    future.set_exception(exc)
                else:
                    self._log_exception(exc)
                    future.set_result(self)
            finally:
                self.subscriptions_map.finished_subscribing()

        asyncio.ensure_future(_async_wrap_subscribe())
        return future

    def _log_exception(self, exc):
        """Log an exception during subscription."""
        msg = (
            "An Exception occurred: {}.".format(exc)
            + " Subscription to {},".format(
                self.service.base_url + self.service.event_subscription_url
            )
            + " sid: {} has been cancelled".format(self.sid)
        )
        log.exception(msg)

    async def renew(
        self, requested_timeout=None, is_autorenew=False, strict=True
    ):  # pylint: disable=invalid-overridden-method
        """renew(requested_timeout=None)
        Renew the event subscription.
        You should not try to renew a subscription which has been
        unsubscribed, or once it has expired.

        This method calls `events_base.SubscriptionBase.renew`.

        Args:
            requested_timeout (int, optional): The period for which a renewal
                request should be made. If None (the default), use the timeout
                requested on subscription.
            is_autorenew (bool, optional): Whether this is an autorenewal.
                Default `False`.
            strict (bool, optional): If True and an Exception occurs during
                execution, the Exception will be raised or, if False, the
                Exception will be logged and the Subscription instance will be
                returned. Default `True`.

        Returns:
            `Subscription`: The Subscription instance.

        """
        try:
            return await super().renew(requested_timeout, is_autorenew)
        except Exception as exc:  # pylint: disable=broad-except
            self._cancel_subscription(exc)
            if self.auto_renew_fail is not None and hasattr(
                self.auto_renew_fail, "__call__"
            ):
                # pylint: disable=not-callable
                self.auto_renew_fail(exc)
            else:
                self._log_exception(exc)
            if strict:
                raise
            return self

    async def unsubscribe(
        self, strict=True
    ):  # pylint: disable=invalid-overridden-method
        """unsubscribe()
        Unsubscribe from the service's events.
        Once unsubscribed, a Subscription instance should not be reused

        This method calls `events_base.SubscriptionBase.unsubscribe`.

        Args:
            strict (bool, optional): If True and an Exception occurs during
                execution, the Exception will be raised or, if False, the
                Exception will be logged and the Subscription instance will be
                returned. Default `True`.

        Returns:
            `Subscription`: The Subscription instance.
        """
        try:
            unsub = super().unsubscribe()
            if unsub is None:
                return
            await unsub
        except Exception as exc:  # pylint: disable=broad-except
            if strict:
                raise
            self._log_exception(exc)
            return self

    def _auto_renew_start(self, interval):
        """Starts the auto_renew loop."""
        self._auto_renew_task = asyncio.get_event_loop().call_later(
            interval, self._auto_renew_run, interval
        )

    def _auto_renew_run(self, interval):
        asyncio.ensure_future(self.renew(is_autorenew=True, strict=False))
        self._auto_renew_start(interval)

    def _auto_renew_cancel(self):
        """Cancels the auto_renew loop"""
        if self._auto_renew_task:
            self._auto_renew_task.cancel()
            self._auto_renew_task = None

    # pylint: disable=no-self-use, too-many-branches, too-many-arguments
    def _request(self, method, url, headers, success, unconditional=None):
        """Sends an HTTP request.

        Args:
            method (str): 'SUBSCRIBE' or 'UNSUBSCRIBE'.
            url (str): The full endpoint to which the request is being sent.
            headers (dict): A dict of headers, each key and each value being
                of type `str`.
            success (function): A function to be called if the
                request succeeds. The function will be called with a dict
                of response headers as its only parameter.
            unconditional (function): An optional function to be called after
                the request is complete, regardless of its success. Takes
                no parameters.

        """

        async def _async_make_request():
            response = await self.event_listener.session.request(
                method, url, headers=headers
            )
            if response.ok:
                success(response.headers)
            if unconditional:
                unconditional()

        return _async_make_request()


class nullcontext:  # pylint: disable=invalid-name
    """Context manager that does no additional processing.

    Backport from python 3.7+ for older pythons.
    """

    def __init__(self, enter_result=None):
        self.enter_result = enter_result

    def __enter__(self):
        return self.enter_result

    def __exit__(self, *excinfo):
        pass


class SubscriptionsMapAio(SubscriptionsMap):
    """Maintains a mapping of sids to `soco.events_asyncio.Subscription`
    instances. Registers each subscription to be unsubscribed at exit.

    Inherits from `soco.events_base.SubscriptionsMap`.
    """

    def __init__(self):
        super().__init__()
        # A counter of calls to Subscription.subscribe
        # that have started but not completed. This is
        # to prevent the event listener from being stopped prematurely
        self._pending = 0
        self.subscriptions_lock = nullcontext()

    def register(self, subscription):
        """Register a subscription by updating local mapping of sid to
        subscription and registering it to be unsubscribed at exit.

        Args:
            subscription(`soco.events_asyncio.Subscription`): the subscription
                to be registered.

        """

        # Add the subscription to the local dict of subscriptions so it
        # can be looked up by sid
        self.subscriptions[subscription.sid] = subscription

    def subscribing(self):
        """Called when the `Subscription.subscribe` method
        commences execution.
        """
        # Increment the counter
        self._pending += 1

    def finished_subscribing(self):
        """Called when the `Subscription.subscribe` method
        completes execution.
        """
        # Decrement the counter
        self._pending -= 1

    @property
    def count(self):
        """
        `int`: The number of active or pending subscriptions.
        """
        return len(self.subscriptions) + self._pending


subscriptions_map = SubscriptionsMapAio()  # pylint: disable=C0103
event_listener = EventListener()  # pylint: disable=C0103



from unittest import mock
import pytest
import requests_mock

from pysonos import SoCo
from pysonos.data_structures import DidlMusicTrack, to_didl_string
from pysonos.exceptions import (
    SoCoSlaveException,
    SoCoUPnPException,
    SoCoNotVisibleException,
    NotSupportedException,
)
from pysonos.groups import ZoneGroup
from pysonos.xml import XML

IP_ADDR = "192.168.1.101"


@pytest.fixture()
def moco():
    """A mock soco with fake services and hardcoded is_coordinator.

    Allows calls to services to be tracked. Should not cause any network
    access
    """
    services = (
        "AVTransport",
        "RenderingControl",
        "DeviceProperties",
        "ContentDirectory",
        "ZoneGroupTopology",
        "GroupRenderingControl",
    )
    patchers = [mock.patch("pysonos.core.{}".format(service)) for service in services]
    for patch in patchers:
        patch.start()
    with mock.patch(
        "pysonos.SoCo.is_coordinator", new_callable=mock.PropertyMock
    ) as is_coord:
        is_coord = True
        yield SoCo(IP_ADDR)
    for patch in reversed(patchers):
        patch.stop()


@pytest.fixture()
def moco_only_on_master():
    """A mock soco with fake services.

    Allows calls to services to be tracked. Should not cause any network
    access
    """
    services = (
        "AVTransport",
        "RenderingControl",
        "DeviceProperties",
        "ContentDirectory",
        "ZoneGroupTopology",
        "GroupRenderingControl",
    )
    patchers = [mock.patch("pysonos.core.{}".format(service)) for service in services]
    for patch in patchers:
        patch.start()
    yield SoCo(IP_ADDR)
    for patch in reversed(patchers):
        patch.stop()


ZGS = (
    """<ZoneGroupState>
      <ZoneGroups>
        <ZoneGroup Coordinator="RINCON_000XXX1400" ID="RINCON_000E5876A0E801400:210">
          <ZoneGroupMember
              UUID="RINCON_000XXX1400"
              Location="http://192.168.1.101:1400/xml/device_description.xml"
              ZoneName="Kitchen"
              Icon="x-rincon-roomicon:kitchen"
              Configuration="1"
              SoftwareVersion="49.2-64250"
              MinCompatibleVersion="48.0-00000"
              LegacyCompatibleVersion="36.0-00000"
              BootSeq="162"
              TVConfigurationError="0"
              HdmiCecAvailable="0"
              WirelessMode="0"
              WirelessLeafOnly="0"
              HasConfiguredSSID="1"
              ChannelFreq="2437"
              BehindWifiExtender="0"
              WifiEnabled="1"
              Orientation="0"
              RoomCalibrationState="4"
              SecureRegState="3"
              VoiceConfigState="0"
              MicEnabled="0"
              AirPlayEnabled="0"
              IdleState="1"
              MoreInfo=""/>
          <ZoneGroupMember
              UUID="RINCON_B8E93781F3EA01400"
              Location="http://192.168.1.31:1400/xml/device_description.xml"
              ZoneName="Living Room"
              Icon="x-rincon-roomicon:masterbedroom"
              Configuration="1"
              SoftwareVersion="49.2-64250"
              MinCompatibleVersion="48.0-00000"
              LegacyCompatibleVersion="36.0-00000"
              BootSeq="226"
              TVConfigurationError="0"
              HdmiCecAvailable="0"
              WirelessMode="0"
              WirelessLeafOnly="0"
              HasConfiguredSSID="1"
              ChannelFreq="2437"
              BehindWifiExtender="0"
              WifiEnabled="1"
              Orientation="0"
              RoomCalibrationState="4"
              SecureRegState="3"
              VoiceConfigState="0"
              MicEnabled="0"
              AirPlayEnabled="0"
              IdleState="1"
              MoreInfo=""/>
        </ZoneGroup>
        <ZoneGroup Coordinator="RINCON_000E58A53FAE01400" """
    """ID="RINCON_000E58A53FAE01400:107">
          <ZoneGroupMember
              UUID="RINCON_000E58A53FAE01400"
              Location="http://192.168.1.173:1400/xml/device_description.xml"
              ZoneName="Stue"
              Icon="x-rincon-roomicon:living"
              Configuration="1"
              SoftwareVersion="49.2-64250"
              MinCompatibleVersion="48.0-00000"
              LegacyCompatibleVersion="36.0-00000"
              BootSeq="1777"
              TVConfigurationError="0"
              HdmiCecAvailable="0"
              WirelessMode="0"
              WirelessLeafOnly="0"
              HasConfiguredSSID="0"
              ChannelFreq="2437"
              BehindWifiExtender="0"
              WifiEnabled="1"
              Orientation="0"
              RoomCalibrationState="4"
              SecureRegState="3"
              VoiceConfigState="0"
              MicEnabled="0"
              AirPlayEnabled="0"
              IdleState="1"
              MoreInfo=""/>
        </ZoneGroup>
        <ZoneGroup Coordinator="RINCON_000E5884455C01400" """
    """ID="RINCON_000E5884455C01400:114">
          <ZoneGroupMember
              UUID="RINCON_000E5884455C01400"
              Location="http://192.168.1.197:1400/xml/device_description.xml"
              ZoneName="Kontor"
              Icon="x-rincon-roomicon:office"
              Configuration="1"
              SoftwareVersion="49.2-64250"
              MinCompatibleVersion="48.0-00000"
              LegacyCompatibleVersion="36.0-00000"
              BootSeq="196"
              TVConfigurationError="0"
              HdmiCecAvailable="0"
              WirelessMode="0"
              WirelessLeafOnly="0"
              HasConfiguredSSID="1"
              ChannelFreq="2437"
              BehindWifiExtender="0"
              WifiEnabled="1"
              Orientation="0"
              RoomCalibrationState="4"
              SecureRegState="3"
              VoiceConfigState="0"
              MicEnabled="0"
              AirPlayEnabled="0"
              IdleState="0"
              MoreInfo=""/>
        </ZoneGroup>
      </ZoneGroups>
      <VanishedDevices></VanishedDevices>
    </ZoneGroupState>"""
)


@pytest.fixture()
def moco_zgs(moco):
    """A mock soco with zone group state."""
    moco.zoneGroupTopology.GetZoneGroupState.return_value = {"ZoneGroupState": ZGS}
    yield moco


class TestSoco:
    device_description = (
        """<?xml version="1.0" encoding="utf-8" ?>
        <root xmlns="urn:schemas-upnp-org:device-1-0">
          <specVersion>
            <major>1</major>
            <minor>0</minor>
          </specVersion>
          <device>
            <deviceType>urn:schemas-upnp-org:device:ZonePlayer:1</deviceType>
            <friendlyName>"""
        + IP_ADDR
        + """ - Sonos Play:5</friendlyName>
            <manufacturer>Sonos, Inc.</manufacturer>
            <manufacturerURL>http://www.sonos.com</manufacturerURL>
            <modelNumber>S3</modelNumber>
            <modelDescription>Sonos Play:5</modelDescription>
            <modelName>Sonos Play:3</modelName>
            <modelURL>http://www.sonos.com/products/zoneplayers/S5</modelURL>
            <softwareVersion>49.2-64250</softwareVersion>
            <hardwareVersion>1.8.1.2-2</hardwareVersion>
            <serialNum>00-11-22-33-44-55:E</serialNum>
            <UDN>uuid:RINCON_000E5884455C01400</UDN>
            <iconList>
              <icon>
                <id>0</id>
                <mimetype>image/png</mimetype>
                <width>48</width>
                <height>48</height>
                <depth>24</depth>
                <url>/img/icon-S3.png</url>
              </icon>
            </iconList>
            <minCompatibleVersion>48.0-00000</minCompatibleVersion>
            <legacyCompatibleVersion>36.0-00000</legacyCompatibleVersion>
            <apiVersion>1.11.1</apiVersion>
            <minApiVersion>1.1.0</minApiVersion>
            <displayVersion>10.1.2</displayVersion>
            <extraVersion>OTP: 1.1.1(1-16-4-zp5s-0.5)</extraVersion>
            <roomName>Room</roomName>
            <displayName>Play:5</displayName>
            <zoneType>5</zoneType>
            <feature1>0x02000002</feature1>
            <feature2>0x00006172</feature2>
            <feature3>0x0003002a</feature3>
            <seriesid>P100</seriesid>
            <variant>0</variant>
            <internalSpeakerSize>3</internalSpeakerSize>
            <bassExtension>0.000</bassExtension>
            <satGainOffset>0.000</satGainOffset>
            <memory>32</memory>
            <flash>32</flash>
            <flashRepartitioned>1</flashRepartitioned>
            <ampOnTime>425</ampOnTime>
            <retailMode>0</retailMode>
            <serviceList>
              <service>
                <serviceType>urn:schemas-upnp-org:service:AlarmClock:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:AlarmClock</serviceId>
                <controlURL>/AlarmClock/Control</controlURL>
                <eventSubURL>/AlarmClock/Event</eventSubURL>
                <SCPDURL>/xml/AlarmClock1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:MusicServices:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:MusicServices</serviceId>
                <controlURL>/MusicServices/Control</controlURL>
                <eventSubURL>/MusicServices/Event</eventSubURL>
                <SCPDURL>/xml/MusicServices1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:AudioIn:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:AudioIn</serviceId>
                <controlURL>/AudioIn/Control</controlURL>
                <eventSubURL>/AudioIn/Event</eventSubURL>
                <SCPDURL>/xml/AudioIn1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:DeviceProperties:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:DeviceProperties</serviceId>
                <controlURL>/DeviceProperties/Control</controlURL>
                <eventSubURL>/DeviceProperties/Event</eventSubURL>
                <SCPDURL>/xml/DeviceProperties1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:SystemProperties:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:SystemProperties</serviceId>
                <controlURL>/SystemProperties/Control</controlURL>
                <eventSubURL>/SystemProperties/Event</eventSubURL>
                <SCPDURL>/xml/SystemProperties1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:ZoneGroupTopology:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:ZoneGroupTopology</serviceId>
                <controlURL>/ZoneGroupTopology/Control</controlURL>
                <eventSubURL>/ZoneGroupTopology/Event</eventSubURL>
                <SCPDURL>/xml/ZoneGroupTopology1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:GroupManagement:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:GroupManagement</serviceId>
                <controlURL>/GroupManagement/Control</controlURL>
                <eventSubURL>/GroupManagement/Event</eventSubURL>
                <SCPDURL>/xml/GroupManagement1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-tencent-com:service:QPlay:1</serviceType>
                <serviceId>urn:tencent-com:serviceId:QPlay</serviceId>
                <controlURL>/QPlay/Control</controlURL>
                <eventSubURL>/QPlay/Event</eventSubURL>
                <SCPDURL>/xml/QPlay1.xml</SCPDURL>
              </service>
            </serviceList>
            <deviceList>
              <device>
          <deviceType>urn:schemas-upnp-org:device:MediaServer:1</deviceType>
          <friendlyName>192.168.1.197 - Sonos Play:5 Media Server</friendlyName>
          <manufacturer>Sonos, Inc.</manufacturer>
          <manufacturerURL>http://www.sonos.com</manufacturerURL>
          <modelNumber>S5</modelNumber>
          <modelDescription>Sonos Play:5 Media Server</modelDescription>
          <modelName>Sonos Play:5</modelName>
          <modelURL>http://www.sonos.com/products/zoneplayers/S5</modelURL>
          <UDN>uuid:RINCON_000E5884455C01400_MS</UDN>
          <serviceList>
            <service>
              <serviceType>urn:schemas-upnp-org:service:ContentDirectory:1</serviceType>
              <serviceId>urn:upnp-org:serviceId:ContentDirectory</serviceId>
              <controlURL>/MediaServer/ContentDirectory/Control</controlURL>
              <eventSubURL>/MediaServer/ContentDirectory/Event</eventSubURL>
              <SCPDURL>/xml/ContentDirectory1.xml</SCPDURL>
            </service>
            <service>
              <serviceType>urn:schemas-upnp-org:service:ConnectionManager:1</serviceType>
                    <serviceId>urn:upnp-org:serviceId:ConnectionManager</serviceId>
                    <controlURL>/MediaServer/ConnectionManager/Control</controlURL>
                    <eventSubURL>/MediaServer/ConnectionManager/Event</eventSubURL>
                    <SCPDURL>/xml/ConnectionManager1.xml</SCPDURL>
                  </service>
                </serviceList>
              </device>
              <device>
                <deviceType>urn:schemas-upnp-org:device:MediaRenderer:1</deviceType>
          <friendlyName>Kontor - Sonos Play:5 Media Renderer</friendlyName>
          <manufacturer>Sonos, Inc.</manufacturer>
          <manufacturerURL>http://www.sonos.com</manufacturerURL>
          <modelNumber>S5</modelNumber>
          <modelDescription>Sonos Play:5 Media Renderer</modelDescription>
          <modelName>Sonos Play:5</modelName>
          <modelURL>http://www.sonos.com/products/zoneplayers/S5</modelURL>
                <UDN>uuid:RINCON_000E5884455C01400_MR</UDN>
                <serviceList>
                  <service>
                    <serviceType>urn:schemas-upnp-org:service:RenderingControl:1</serviceType>
                    <serviceId>urn:upnp-org:serviceId:RenderingControl</serviceId>
                    <controlURL>/MediaRenderer/RenderingControl/Control</controlURL>
                    <eventSubURL>/MediaRenderer/RenderingControl/Event</eventSubURL>
                    <SCPDURL>/xml/RenderingControl1.xml</SCPDURL>
                  </service>
                  <service>
                    <serviceType>urn:schemas-upnp-org:service:ConnectionManager:1</serviceType>
                    <serviceId>urn:upnp-org:serviceId:ConnectionManager</serviceId>
                    <controlURL>/MediaRenderer/ConnectionManager/Control</controlURL>
                    <eventSubURL>/MediaRenderer/ConnectionManager/Event</eventSubURL>
                    <SCPDURL>/xml/ConnectionManager1.xml</SCPDURL>
                  </service>
                  <service>
                    <serviceType>urn:schemas-upnp-org:service:AVTransport:1</serviceType>
                    <serviceId>urn:upnp-org:serviceId:AVTransport</serviceId>
                    <controlURL>/MediaRenderer/AVTransport/Control</controlURL>
                    <eventSubURL>/MediaRenderer/AVTransport/Event</eventSubURL>
                    <SCPDURL>/xml/AVTransport1.xml</SCPDURL>
                  </service>
                  <service>
                    <serviceType>urn:schemas-sonos-com:service:Queue:1</serviceType>
                    <serviceId>urn:sonos-com:serviceId:Queue</serviceId>
                    <controlURL>/MediaRenderer/Queue/Control</controlURL>
                    <eventSubURL>/MediaRenderer/Queue/Event</eventSubURL>
                    <SCPDURL>/xml/Queue1.xml</SCPDURL>
                  </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:GroupRenderingControl:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:GroupRenderingControl</serviceId>
                <controlURL>/MediaRenderer/GroupRenderingControl/Control</controlURL>
                <eventSubURL>/MediaRenderer/GroupRenderingControl/Event</eventSubURL>
                <SCPDURL>/xml/GroupRenderingControl1.xml</SCPDURL>
              </service>
                </serviceList>
                <X_Rhapsody-Extension """
        """xmlns="http://www.real.com/rhapsody/xmlns/upnp-1-0">
                  <deviceID>urn:rhapsody-real-com:device-id-1-0:"""
        """sonos_1:RINCON_000E5884455C01400</deviceID>
                    <deviceCapabilities>
                      <interactionPattern type="real-rhapsody-upnp-1-0"/>
                    </deviceCapabilities>
                </X_Rhapsody-Extension>
                <qq:X_QPlay_SoftwareCapability xmlns:qq="http://www.tencent.com">"""
        """QPlay:2</qq:X_QPlay_SoftwareCapability>
                <iconList>
                  <icon>
                    <mimetype>image/png</mimetype>
                    <width>48</width>
                    <height>48</height>
                    <depth>24</depth>
                    <url>/img/icon-S5.png</url>
                  </icon>
                </iconList>
              </device>
            </deviceList>
          </device>
        </root>
    """
    )

    @pytest.mark.parametrize("bad_ip_addr", ["not_ip", "555.555.555.555"])
    def test_soco_bad_ip(self, bad_ip_addr):
        with pytest.raises(ValueError):
            speaker = SoCo(bad_ip_addr)

    def test_soco_init(self, moco):
        assert moco.ip_address == IP_ADDR
        assert moco.speaker_info == {}

    def test_soco_str(self, moco):
        assert str(moco) == "<SoCo object at ip {}>".format(IP_ADDR)

    def test_soco_repr(self, moco):
        assert repr(moco) == 'SoCo("{}")'.format(IP_ADDR)

    @pytest.mark.parametrize(
        "model_name",
        (
            ("Play:5", False),
            ("Sonos One", False),
            ("PLAYBAR", True),
            ("Sonos Beam", True),
            ("Sonos Playbar", True),
            ("Sonos Playbase", True),
            ("Sonos Arc", True),
            ("Sonos Arc SL", True),
        ),
    )
    def test_soco_is_soundbar(self, moco, model_name):
        moco._is_soundbar = None
        moco.speaker_info["model_name"] = model_name[0]
        assert moco.is_soundbar == model_name[1]

    @mock.patch("pysonos.core.requests")
    @pytest.mark.parametrize("refresh", [None, False, True])
    def test_soco_get_speaker_info_speaker_not_set_refresh(
        self, mocr, moco_zgs, refresh
    ):
        """Internal speaker_info not set; Refresh all values (default, False,
        True)

        => should update
        """
        response = mock.MagicMock()
        mocr.get.return_value = response
        response.content = self.device_description
        # save old state
        old = moco_zgs.speaker_info
        moco_zgs.speaker_info = {}
        if refresh is None:
            res = moco_zgs.get_speaker_info()
        else:
            res = moco_zgs.get_speaker_info(refresh)
        # restore original value
        moco_zgs.speaker_info = old
        mocr.get.assert_called_once_with(
            "http://" + IP_ADDR + ":1400/xml/device_description.xml",
            timeout=None,
        )
        should = {
            "zone_name": "Room",
            "player_icon": "/img/icon-S3.png",
            "uid": "RINCON_000XXX1400",
            "serial_number": "00-11-22-33-44-55:E",
            "software_version": "49.2-64250",
            "hardware_version": "1.8.1.2-2",
            "model_number": "S3",
            "model_name": "Sonos Play:3",
            "display_version": "10.1.2",
            "mac_address": "00-11-22-33-44-55",
        }
        assert should == res

    @mock.patch("pysonos.core.requests")
    @pytest.mark.parametrize("refresh", [None, False])
    def test_soco_get_speaker_info_speaker_set_no_refresh(
        self, mocr, moco_zgs, refresh
    ):
        """Internal speaker_info set; No refresh (default, False)

        => should not update
        """
        should = {"info": "yes"}
        # save old state
        old = moco_zgs.speaker_info
        moco_zgs.speaker_info = should
        if refresh is None:
            res = moco_zgs.get_speaker_info()
        else:
            res = moco_zgs.get_speaker_info(refresh)
        # restore original value
        moco_zgs.speaker_info = old
        # got 'should' returned
        assert res is should
        # no network request performed
        assert not mocr.get.called

    @mock.patch("pysonos.core.requests")
    @pytest.mark.parametrize("should", [{}, {"info": "yes"}])
    def test_soco_get_speaker_info_speaker_set_no_refresh(self, mocr, moco_zgs, should):
        """Internal speaker_info not set/set; Refresh True.

        => should update
        """
        response = mock.MagicMock()
        mocr.get.return_value = response
        response.content = self.device_description
        # save old state
        old = moco_zgs.speaker_info
        moco_zgs.speaker_info = should
        res = moco_zgs.get_speaker_info(True)
        # restore original value
        moco_zgs.speaker_info = old
        mocr.get.assert_called_once_with(
            "http://" + IP_ADDR + ":1400/xml/device_description.xml",
            timeout=None,
        )
        # get_speaker_info only updates internal speaker_info and does not
        # replace it
        should.update(
            {
                "zone_name": "Room",
                "player_icon": "/img/icon-S3.png",
                "uid": "RINCON_000XXX1400",
                "serial_number": "00-11-22-33-44-55:E",
                "software_version": "29.5-91030",
                "hardware_version": "1.8.1.2-2",
                "model_number": "S3",
                "model_name": "Sonos Play:3",
                "display_version": "10.1.2",
                "mac_address": "00-11-22-33-44-55",
            }
        )
        assert should == res


class TestAVTransport:
    @pytest.mark.parametrize(
        "playmode",
        [
            "NORMAL",
            "SHUFFLE_NOREPEAT",
            "SHUFFLE",
            "REPEAT_ALL",
            "SHUFFLE_REPEAT_ONE",
            "REPEAT_ONE",
        ],
    )
    def test_soco_play_mode_values(self, moco, playmode):
        moco.avTransport.GetTransportSettings.return_value = {"PlayMode": playmode}
        assert moco.play_mode == playmode
        moco.avTransport.GetTransportSettings.assert_called_once_with(
            [("InstanceID", 0)]
        )
        moco.avTransport.reset_mock()

    def test_soco_play_mode_bad_value(self, moco):
        with pytest.raises(KeyError):
            moco.play_mode = "BAD_VALUE"
        assert not moco.avTransport.SetPlayMode.called

    def test_soco_play_mode_lowercase(self, moco):
        moco.play_mode = "normal"
        moco.avTransport.SetPlayMode.assert_called_once_with(
            [("InstanceID", 0), ("NewPlayMode", "NORMAL")]
        )

    def test_available_actions(self, moco):
        moco.avTransport.GetCurrentTransportActions.return_value = {
            "Actions": "Set, Stop, Pause, Play, X_DLNA_SeekTime, X_DLNA_SeekTrackNr"
        }
        assert moco.available_actions == [
            "Set",
            "Stop",
            "Pause",
            "Play",
            "SeekTime",
            "SeekTrackNr",
        ]
        moco.avTransport.GetCurrentTransportActions.assert_called_once_with(
            [("InstanceID", 0)]
        )

    def test_soco_cross_fade2(self, moco):
        moco.avTransport.GetCrossfadeMode.return_value = {"CrossfadeMode": "1"}
        assert moco.cross_fade
        moco.avTransport.GetCrossfadeMode.return_value = {"CrossfadeMode": "0"}
        assert not moco.cross_fade
        moco.avTransport.GetCrossfadeMode.assert_called_with([("InstanceID", 0)])
        moco.cross_fade = True
        moco.avTransport.SetCrossfadeMode.assert_called_once_with(
            [("InstanceID", 0), ("CrossfadeMode", "1")]
        )

    def test_soco_play(self, moco):
        moco.play()
        moco.avTransport.Play.assert_called_once_with([("InstanceID", 0), ("Speed", 1)])

    # Test that uris are forced to Radio style display and controls when
    # force_radio is True prefix is replaced with "x-rincon-mp3radion://"
    # first set of test with no forcing, second set with force_radio=True

    # No forcing
    @pytest.mark.parametrize(
        "uri_in, uri_passed",
        [
            (
                "x-file-cifs://server/MyNiceRing.mp3",
                "x-file-cifs://server/MyNiceRing.mp3",
            ),
            (
                "http://archive.org/download/TenD2005-07-16t_64kb.mp3",
                "http://archive.org/download/TenD2005-07-16t_64kb.mp3",
            ),
            (
                "x-sonosapi-radio:station%3a%3aps.147077612?sid=203&flags=76&sn=2",
                "x-sonosapi-radio:station%3a%3aps.147077612?sid=203&flags=76&sn=2",
            ),
        ],
    )
    def test_soco_play_uri(self, moco, uri_in, uri_passed):
        moco.play_uri(uri_in)
        moco.avTransport.SetAVTransportURI.assert_called_once_with(
            [("InstanceID", 0), ("CurrentURI", uri_passed), ("CurrentURIMetaData", "")]
        )
        moco.avTransport.reset_mock()

    # with force_radio=True
    @pytest.mark.parametrize(
        "uri_in, uri_passed",
        [
            (
                "http://archive.org/download/TenD2005-07-16t_64kb.mp3",
                "x-rincon-mp3radio://archive.org/download/TenD2005-07-16t_64kb.mp3",
            ),
            (
                "https://archive.org/download/TenD2005-07-16t_64kb.mp3",
                "x-rincon-mp3radio://archive.org/download/TenD2005-07-16t_64kb.mp3",
            ),
            (
                "aac://icy-e-bz-04-cr.sharp-stream.com/magic1054.aac?amsparams="
                "playerid:BMUK_tunein;skey:1483570441&awparams=loggedin:false",
                "x-rincon-mp3radio://icy-e-bz-04-cr.sharp-stream.com/magic1054.aac?"
                "amsparams=playerid:BMUK_tunein;skey:1483570441&awparams=loggedin:"
                "false",
            ),
        ],
    )
    def test_soco_play_uri_force_radio(self, moco, uri_in, uri_passed):
        moco.play_uri(uri_in, force_radio=True)
        moco.avTransport.SetAVTransportURI.assert_called_once_with(
            [("InstanceID", 0), ("CurrentURI", uri_passed), ("CurrentURIMetaData", "")]
        )
        moco.avTransport.reset_mock()

    def test_soco_play_uri_calls_play(self, moco):
        uri = (
            "http://archive.org/download/tend2005-07-16.flac16/"
            "tend2005-07-16t10wonderboy_64kb.mp3"
        )
        moco.play_uri(uri)

        moco.avTransport.Play.assert_called_with([("InstanceID", 0), ("Speed", 1)])

    def test_soco_play_uri_with_title(self, moco):
        uri = (
            "http://archive.org/download/tend2005-07-16.flac16/"
            "tend2005-07-16t10wonderboy_64kb.mp3"
        )
        moco.play_uri(uri, title="<Fast & Loose>")

        moco.avTransport.SetAVTransportURI.assert_called_with(
            [
                ("InstanceID", 0),
                ("CurrentURI", uri),
                (
                    "CurrentURIMetaData",
                    '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                    'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                    'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                    'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">'
                    '<item id="R:0/0/0" parentID="R:0/0" restricted="true">'
                    "<dc:title>&lt;Fast &amp; Loose&gt;</dc:title>"
                    "<upnp:class>object.item.audioItem.audioBroadcast</upnp:class>"
                    '<desc id="cdudn" nameSpace="urn:schemas-rinconnetworks-com:'
                    'metadata-1-0/">SA_RINCON65031_</desc></item></DIDL-Lite>',
                ),
            ]
        )

    def test_soco_pause(self, moco):
        moco.pause()
        moco.avTransport.Pause.assert_called_once_with(
            [("InstanceID", 0), ("Speed", 1)]
        )

    def test_soco_stop(self, moco):
        moco.stop()
        moco.avTransport.Stop.assert_called_once_with([("InstanceID", 0), ("Speed", 1)])

    def test_soco_end_direct_control_session(self, moco):
        moco.end_direct_control_session()
        moco.avTransport.EndDirectControlSession.assert_called_once_with(
            [("InstanceID", 0)]
        )

    def test_soco_next(self, moco):
        moco.next()
        moco.avTransport.Next.assert_called_once_with([("InstanceID", 0), ("Speed", 1)])

    def test_soco_previous(self, moco):
        moco.previous()
        moco.avTransport.Previous.assert_called_once_with(
            [("InstanceID", 0), ("Speed", 1)]
        )

    @pytest.mark.parametrize(
        "bad_timestamp", ["NOT_VALID", "12:34:56:78", "99", "12:3:4"]
    )
    def test_soco_seek_invalid(self, moco, bad_timestamp):
        with pytest.raises(ValueError):
            moco.seek(bad_timestamp)
        assert not moco.avTransport.Seek.called

    @pytest.mark.parametrize(
        "timestamp",
        [
            # '12:34', # Should this be valid?
            "1:23:45",
            "10:23:45",
            "12:78:78",  # Should this really be valid?
        ],
    )
    def test_soco_seek_valid(self, moco, timestamp):
        moco.seek(timestamp)
        moco.avTransport.Seek.assert_called_once_with(
            [("InstanceID", 0), ("Unit", "REL_TIME"), ("Target", timestamp)]
        )
        moco.avTransport.reset_mock()

    def test_soco_current_transport_info(self, moco):
        moco.avTransport.GetTransportInfo.return_value = {
            "CurrentTransportState": "PLAYING",
            "CurrentTransportStatus": "OK",
            "CurrentSpeed": "1",
        }
        playstate = moco.get_current_transport_info()
        moco.avTransport.GetTransportInfo.assert_called_once_with([("InstanceID", 0)])
        assert playstate["current_transport_state"] == "PLAYING"
        assert playstate["current_transport_status"] == "OK"
        assert playstate["current_transport_speed"] == "1"

    def test_soco_get_queue(self, moco):
        moco.contentDirectory.Browse.return_value = {
            "Result": (
                '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">'
                '<item id="Q:0/1" parentID="Q:0" restricted="true">'
                '<res protocolInfo="fake.com-fake-direct:*:audio/mp3:*" '
                'duration="0:02:32">radea:Tra.12345678.mp3</res>'
                "<upnp:albumArtURI>/getaa?r=1&amp;u=radea%3aTra.12345678.mp3"
                "</upnp:albumArtURI><dc:title>Item 1 Title</dc:title>"
                "<upnp:class>object.item.audioItem.musicTrack</upnp:class>"
                "<dc:creator>Item 1 Creator</dc:creator>"
                "<upnp:album>Item 1 Album</upnp:album></item></DIDL-Lite>"
            ),
            "NumberReturned": "1",
            "TotalMatches": "10",
            "UpdateID": "1",
        }
        queue = moco.get_queue(start=8, max_items=32)
        moco.contentDirectory.Browse.assert_called_once_with(
            [
                ("ObjectID", "Q:0"),
                ("BrowseFlag", "BrowseDirectChildren"),
                ("Filter", "*"),
                ("StartingIndex", 8),
                ("RequestedCount", 32),
                ("SortCriteria", ""),
            ]
        )
        assert queue is not None
        assert len(queue) == 1
        moco.contentDirectory.reset_mock()

    def test_soco_queue_size(self, moco):
        moco.contentDirectory.Browse.return_value = {
            "NumberReturned": "1",
            "Result": (
                '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">'
                '<container id="Q:0" parentID="Q:" restricted="true" childCount="384">'
                "<dc:title>Queue Instance 0</dc:title><upnp:class>"
                "object.container.playlistContainer</upnp:class>"
                '<res protocolInfo="x-rincon-queue:*:*:*">'
                "x-rincon-queue:RINCON_00012345678901400#0</res>"
                "</container></DIDL-Lite>"
            ),
            "TotalMatches": "1",
            "UpdateID": "1",
        }
        queue_size = moco.queue_size
        moco.contentDirectory.Browse.assert_called_once_with(
            [
                ("ObjectID", "Q:0"),
                ("BrowseFlag", "BrowseMetadata"),
                ("Filter", "*"),
                ("StartingIndex", 0),
                ("RequestedCount", 1),
                ("SortCriteria", ""),
            ]
        )
        assert queue_size == 384
        moco.contentDirectory.reset_mock()

    def test_join(self, moco_zgs):
        moco2 = mock.Mock()
        moco2.uid = "RINCON_000XXX1400"
        moco_zgs.join(moco2)
        moco_zgs.avTransport.SetAVTransportURI.assert_called_once_with(
            [
                ("InstanceID", 0),
                ("CurrentURI", "x-rincon:RINCON_000XXX1400"),
                ("CurrentURIMetaData", ""),
            ]
        )

    def test_unjoin(self, moco_zgs):
        moco_zgs.unjoin()
        moco_zgs.avTransport.BecomeCoordinatorOfStandaloneGroup.assert_called_once_with(
            [("InstanceID", 0)]
        )

    def test_switch_to_line_in(self, moco_zgs):
        moco_zgs.avTransport.reset_mock()
        moco_zgs.switch_to_line_in()
        moco_zgs.avTransport.SetAVTransportURI.assert_called_once_with(
            [
                ("InstanceID", 0),
                ("CurrentURI", "x-rincon-stream:RINCON_000XXX1400"),
                ("CurrentURIMetaData", ""),
            ]
        )

        moco_zgs.avTransport.reset_mock()
        moco2 = mock.Mock()
        moco2.uid = "RINCON_000YYY1400"
        moco_zgs.switch_to_line_in(moco2)
        moco_zgs.avTransport.SetAVTransportURI.assert_called_once_with(
            [
                ("InstanceID", 0),
                ("CurrentURI", "x-rincon-stream:RINCON_000YYY1400"),
                ("CurrentURIMetaData", ""),
            ]
        )

    def test_switch_to_tv(self, moco_zgs):
        moco_zgs.avTransport.reset_mock()
        moco_zgs.switch_to_tv()
        moco_zgs.avTransport.SetAVTransportURI.assert_called_once_with(
            [
                ("InstanceID", 0),
                ("CurrentURI", "x-sonos-htastream:RINCON_000XXX1400:spdif"),
                ("CurrentURIMetaData", ""),
            ]
        )

    def test_is_playing_tv(self, moco):
        moco.avTransport.reset_mock()
        moco.avTransport.GetPositionInfo.return_value = {
            "TrackURI": "x-sonos-htastream:RINCON_00012345678901234:spdif"
        }
        playing_tv = moco.is_playing_tv
        assert playing_tv
        moco.avTransport.GetPositionInfo.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master")]
        )

        moco.avTransport.GetPositionInfo.return_value = {
            "TrackURI": "not-tv",
        }
        playing_tv = moco.is_playing_tv
        assert not playing_tv

    def test_is_playing_radio(self, moco):
        moco.avTransport.reset_mock()
        moco.avTransport.GetPositionInfo.return_value = {
            "TrackURI": "x-rincon-mp3radio://example.com:80/myradio"
        }
        playing_tv = moco.is_playing_radio
        assert playing_tv
        moco.avTransport.GetPositionInfo.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master")]
        )

        moco.avTransport.GetPositionInfo.return_value = {
            "TrackURI": "not-radio",
        }
        playing_tv = moco.is_playing_radio
        assert not playing_tv

    def test_is_playing_line_in(self, moco):
        moco.avTransport.reset_mock()
        moco.avTransport.GetPositionInfo.return_value = {
            "TrackURI": "x-rincon-stream:blah-blah"
        }
        playing_tv = moco.is_playing_line_in
        assert playing_tv
        moco.avTransport.GetPositionInfo.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master")]
        )

        moco.avTransport.GetPositionInfo.return_value = {
            "TrackURI": "not-line-in",
        }
        playing_tv = moco.is_playing_line_in
        assert not playing_tv

    def test_create_sonos_playlist(self, moco):
        playlist_name = "cool music"
        playlist_id = 1
        moco.avTransport.CreateSavedQueue.return_value = {
            "AssignedObjectID": "SQ:{}".format(playlist_id)
        }
        playlist = moco.create_sonos_playlist(playlist_name)
        moco.avTransport.CreateSavedQueue.assert_called_once_with(
            [
                ("InstanceID", 0),
                ("Title", playlist_name),
                ("EnqueuedURI", ""),
                ("EnqueuedURIMetaData", ""),
            ]
        )
        assert playlist.title == playlist_name
        expected_uri = "file:///jffs/settings/savedqueues.rsq#{}".format(playlist_id)
        assert playlist.resources[0].uri == expected_uri
        assert playlist.parent_id == "SQ:"

    def test_create_sonos_playlist_from_queue(self, moco):
        playlist_name = "saved queue"
        playlist_id = 1
        moco.avTransport.SaveQueue.return_value = {
            "AssignedObjectID": "SQ:{}".format(playlist_id)
        }
        playlist = moco.create_sonos_playlist_from_queue(playlist_name)
        moco.avTransport.SaveQueue.assert_called_once_with(
            [("InstanceID", 0), ("Title", playlist_name), ("ObjectID", "")]
        )
        assert playlist.title == playlist_name
        expected_uri = "file:///jffs/settings/savedqueues.rsq#{}".format(playlist_id)
        assert playlist.resources[0].uri == expected_uri
        assert playlist.parent_id == "SQ:"

    def test_add_item_to_sonos_playlist(self, moco):
        moco.contentDirectory.reset_mock()

        playlist = mock.Mock()
        playlist.item_id = 7

        ressource = mock.Mock()
        ressource.uri = "fake_uri"
        track = mock.Mock()
        track.resources = [ressource]
        track.uri = "fake_uri"
        track.to_element.return_value = XML.Element("a")

        update_id = 100

        moco.contentDirectory.Browse.return_value = {
            "NumberReturned": "0",
            "Result": (
                '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"></DIDL-Lite>'
            ),
            "TotalMatches": "0",
            "UpdateID": update_id,
        }

        moco.add_item_to_sonos_playlist(track, playlist)
        moco.contentDirectory.Browse.assert_called_once_with(
            [
                ("ObjectID", playlist.item_id),
                ("BrowseFlag", "BrowseDirectChildren"),
                ("Filter", "*"),
                ("StartingIndex", 0),
                ("RequestedCount", 1),
                ("SortCriteria", ""),
            ]
        )
        moco.avTransport.AddURIToSavedQueue.assert_called_once_with(
            [
                ("InstanceID", 0),
                ("UpdateID", update_id),
                ("ObjectID", playlist.item_id),
                ("EnqueuedURI", track.uri),
                ("EnqueuedURIMetaData", to_didl_string(track)),
                ("AddAtIndex", 4294967295),
            ]
        )

    def test_soco_cross_fade(self, moco):
        moco.avTransport.GetCrossfadeMode.return_value = {"CrossfadeMode": "1"}
        assert moco.cross_fade
        moco.avTransport.GetCrossfadeMode.assert_called_once_with([("InstanceID", 0)])
        moco.cross_fade = False
        moco.avTransport.SetCrossfadeMode.assert_called_once_with(
            [("InstanceID", 0), ("CrossfadeMode", "0")]
        )

    def test_shuffle(self, moco):
        moco.avTransport.GetTransportSettings.return_value = {"PlayMode": "NORMAL"}
        assert moco.shuffle is False
        moco.avTransport.GetTransportSettings.assert_called_once_with(
            [("InstanceID", 0)]
        )
        moco.shuffle = True
        moco.avTransport.SetPlayMode.assert_called_once_with(
            [("InstanceID", 0), ("NewPlayMode", "SHUFFLE_NOREPEAT")]
        )

        moco.avTransport.GetTransportSettings.return_value = {
            "PlayMode": "SHUFFLE_NOREPEAT"
        }
        assert moco.shuffle is True
        moco.avTransport.GetTransportSettings.assert_called_with([("InstanceID", 0)])
        moco.shuffle = False
        moco.avTransport.SetPlayMode.assert_called_with(
            [("InstanceID", 0), ("NewPlayMode", "NORMAL")]
        )

    def test_repeat(self, moco):
        moco.avTransport.GetTransportSettings.return_value = {"PlayMode": "NORMAL"}
        assert moco.repeat is False
        moco.avTransport.GetTransportSettings.assert_called_with([("InstanceID", 0)])
        moco.repeat = True
        moco.avTransport.SetPlayMode.assert_called_with(
            [("InstanceID", 0), ("NewPlayMode", "REPEAT_ALL")]
        )

        moco.avTransport.GetTransportSettings.return_value = {"PlayMode": "REPEAT_ALL"}
        assert moco.repeat is True
        moco.avTransport.GetTransportSettings.assert_called_with([("InstanceID", 0)])
        moco.repeat = False
        moco.avTransport.SetPlayMode.assert_called_with(
            [("InstanceID", 0), ("NewPlayMode", "NORMAL")]
        )

        moco.avTransport.GetTransportSettings.return_value = {"PlayMode": "REPEAT_ONE"}
        assert moco.repeat == "ONE"
        moco.avTransport.GetTransportSettings.assert_called_with([("InstanceID", 0)])
        moco.repeat = "ONE"
        moco.avTransport.SetPlayMode.assert_called_with(
            [("InstanceID", 0), ("NewPlayMode", "REPEAT_ONE")]
        )

    def test_set_sleep_timer(self, moco):
        moco.avTransport.reset_mock()
        moco.avTransport.ConfigureSleepTimer.return_value = None
        result = moco.set_sleep_timer(None)
        assert result is None
        moco.avTransport.ConfigureSleepTimer.assert_called_once_with(
            [("InstanceID", 0), ("NewSleepTimerDuration", "")]
        )

        moco.avTransport.reset_mock()
        moco.avTransport.ConfigureSleepTimer.return_value = None
        result = moco.set_sleep_timer(7200)
        assert result is None
        moco.avTransport.ConfigureSleepTimer.assert_called_once_with(
            [("InstanceID", 0), ("NewSleepTimerDuration", "2:00:00")]
        )

        moco.avTransport.reset_mock()
        moco.avTransport.ConfigureSleepTimer.return_value = None
        result = moco.set_sleep_timer(0)
        assert result is None
        moco.avTransport.ConfigureSleepTimer.assert_called_once_with(
            [("InstanceID", 0), ("NewSleepTimerDuration", "0:00:00")]
        )

    @pytest.mark.parametrize("bad_sleep_time", ["BadTime", "00:43:23", "4200s", ""])
    def test_set_sleep_timer_bad_sleep_time(self, moco, bad_sleep_time):
        with pytest.raises(ValueError):
            result = moco.set_sleep_timer(bad_sleep_time)

    def test_get_sleep_timer(self, moco):
        moco.avTransport.reset_mock()
        moco.avTransport.GetRemainingSleepTimerDuration.return_value = {
            "RemainingSleepTimerDuration": "02:00:00",
            "CurrentSleepTimerGeneration": "3",
        }
        result = moco.get_sleep_timer()
        assert result == 7200

        moco.avTransport.reset_mock()
        moco.avTransport.GetRemainingSleepTimerDuration.return_value = {
            "RemainingSleepTimerDuration": "",
            "CurrentSleepTimerGeneration": "0",
        }
        result = moco.get_sleep_timer()
        assert result is None


class TestContentDirectory:
    def test_remove_sonos_playlist_success(self, moco):
        moco.contentDirectory.reset_mock()
        moco.contentDirectory.return_value = True
        result = moco.remove_sonos_playlist("SQ:10")
        moco.contentDirectory.DestroyObject.assert_called_once_with(
            [("ObjectID", "SQ:10")]
        )
        assert result


class TestRenderingControl:
    def test_soco_mute(self, moco):
        moco.renderingControl.GetMute.return_value = {"CurrentMute": "1"}
        assert moco.mute
        moco.renderingControl.reset_mock()
        moco.renderingControl.GetMute.return_value = {"CurrentMute": "0"}
        assert not moco.mute
        moco.renderingControl.GetMute.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master")]
        )
        moco.mute = False
        moco.renderingControl.SetMute.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master"), ("DesiredMute", "0")]
        )
        moco.renderingControl.reset_mock()
        moco.mute = True
        moco.renderingControl.SetMute.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master"), ("DesiredMute", "1")]
        )

    @pytest.mark.parametrize("volume", [1, 4, 10, 100])
    def test_soco_volume(self, moco, volume):
        moco.renderingControl.GetVolume.return_value = {"CurrentVolume": volume}
        assert moco.volume == volume
        moco.renderingControl.GetVolume.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master")]
        )
        moco.renderingControl.reset_mock()

    @pytest.mark.parametrize(
        "vol_set, vol_called",
        [
            (-120, 0),
            (-10, 0),
            (0, 0),
            (5, 5),
            (99, 99),
            (100, 100),
            (110, 100),
            (300, 100),
        ],
    )
    def soco_volume_set(self, moco, vol_set, vol_called):
        moco.volume = vol_set
        moco.renderingControl.SetVolume.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master"), ("DesiredVolume", vol_called)]
        )

    def test_soco_ramp_to_volume(self, moco):
        moco.renderingControl.RampToVolume.return_value = {"RampTime": "12"}
        ramp_time = moco.ramp_to_volume(15)
        moco.renderingControl.RampToVolume.assert_called_once_with(
            [
                ("InstanceID", 0),
                ("Channel", "Master"),
                ("RampType", "SLEEP_TIMER_RAMP_TYPE"),
                ("DesiredVolume", 15),
                ("ResetVolumeAfter", False),
                ("ProgramURI", ""),
            ]
        )
        assert ramp_time == 12

    def test_set_relative_volume(self, moco):
        moco.renderingControl.SetRelativeVolume.return_value = {"NewVolume": "75"}
        new_volume = moco.set_relative_volume(25)
        moco.renderingControl.SetRelativeVolume.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master"), ("Adjustment", 25)]
        )
        assert new_volume == 75

    def test_soco_treble(self, moco):
        moco.renderingControl.GetTreble.return_value = {"CurrentTreble": "15"}
        assert moco.treble == 15
        moco.renderingControl.GetTreble.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master")]
        )
        moco.treble = "10"
        moco.renderingControl.SetTreble.assert_called_once_with(
            [("InstanceID", 0), ("DesiredTreble", 10)]
        )

    def test_soco_loudness(self, moco):
        moco.renderingControl.GetLoudness.return_value = {"CurrentLoudness": "1"}
        assert moco.loudness
        moco.renderingControl.GetLoudness.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master")]
        )
        moco.loudness = False
        moco.renderingControl.SetLoudness.assert_called_once_with(
            [("InstanceID", 0), ("Channel", "Master"), ("DesiredLoudness", "0")]
        )

    def test_soco_trueplay(self, moco):
        moco.renderingControl.GetRoomCalibrationStatus.return_value = {
            "RoomCalibrationAvailable": "0",
            "RoomCalibrationEnabled": "0",
        }
        assert moco.trueplay is None
        moco.renderingControl.GetRoomCalibrationStatus.assert_called_with(
            [("InstanceID", 0)]
        )
        moco.renderingControl.GetRoomCalibrationStatus.return_value = {
            "RoomCalibrationAvailable": "1",
            "RoomCalibrationEnabled": "1",
        }
        assert moco.trueplay
        moco.renderingControl.GetRoomCalibrationStatus.assert_called_with(
            [("InstanceID", 0)]
        )
        # Setter tests for 'is_visible' property, so this needs to be
        # mocked.
        with mock.patch(
            "pysonos.SoCo.is_visible", new_callable=mock.PropertyMock
        ) as mock_is_visible:
            mock_is_visible.return_value = True
            moco.trueplay = False
            moco.renderingControl.SetRoomCalibrationStatus.assert_called_with(
                [
                    ("InstanceID", 0),
                    ("RoomCalibrationEnabled", "0"),
                ]
            )
            moco.trueplay = True
            moco.renderingControl.SetRoomCalibrationStatus.assert_called_with(
                [
                    ("InstanceID", 0),
                    ("RoomCalibrationEnabled", "1"),
                ]
            )
            # Check for exception if attempt to set the property on a
            # non-visible speaker.
            mock_is_visible.return_value = False
            with pytest.raises(SoCoNotVisibleException):
                moco.trueplay = True

    def test_soco_fixed_volume(self, moco):
        moco.renderingControl.GetSupportsOutputFixed.return_value = {
            "CurrentSupportsFixed": "1"
        }
        assert moco.supports_fixed_volume
        moco.renderingControl.GetSupportsOutputFixed.assert_called_with(
            [("InstanceID", 0)]
        )
        moco.renderingControl.GetSupportsOutputFixed.return_value = {
            "CurrentSupportsFixed": "0"
        }
        assert not moco.supports_fixed_volume
        moco.renderingControl.GetSupportsOutputFixed.assert_called_with(
            [("InstanceID", 0)]
        )
        moco.renderingControl.GetOutputFixed.return_value = {"CurrentFixed": "1"}
        assert moco.fixed_volume
        moco.renderingControl.GetOutputFixed.assert_called_once_with(
            [("InstanceID", 0)]
        )
        moco.fixed_volume = False
        moco.renderingControl.SetOutputFixed.assert_called_once_with(
            [("InstanceID", 0), ("DesiredFixed", "0")]
        )
        moco.renderingControl.SetOutputFixed.side_effect = SoCoUPnPException(
            None, None, None
        )
        with pytest.raises(NotSupportedException):
            moco.fixed_volume = True

    def test_soco_balance(self, moco):
        # GetVolume is called twice, once for each of the left
        # and right channels
        moco.renderingControl.GetVolume.return_value = {"CurrentVolume": "100"}
        assert moco.balance == (100, 100)
        moco.renderingControl.GetVolume.assert_any_call(
            [("InstanceID", 0), ("Channel", "LF")]
        )
        moco.renderingControl.GetVolume.assert_any_call(
            [
                ("InstanceID", 0),
                ("Channel", "RF"),
            ]
        )
        # SetVolume is called twice, once for each of the left
        # and right channels
        moco.balance = (0, 100)
        moco.renderingControl.SetVolume.assert_any_call(
            [("InstanceID", 0), ("Channel", "LF"), ("DesiredVolume", 0)]
        )
        moco.renderingControl.SetVolume.assert_any_call(
            [("InstanceID", 0), ("Channel", "RF"), ("DesiredVolume", 100)]
        )


class TestDeviceProperties:
    def test_soco_status_light(self, moco):
        moco.deviceProperties.GetLEDState.return_value = {"CurrentLEDState": "On"}
        assert moco.status_light
        moco.deviceProperties.GetLEDState.return_value = {"CurrentLEDState": "Off"}
        assert not moco.status_light
        moco.deviceProperties.GetLEDState.assert_called_with()
        moco.status_light = False
        moco.deviceProperties.SetLEDState.assert_called_once_with(
            [("DesiredLEDState", "Off")]
        )
        moco.status_light = True
        moco.deviceProperties.SetLEDState.assert_called_with(
            [("DesiredLEDState", "On")]
        )

    def test_buttons_enabled(self, moco):
        moco.deviceProperties.GetButtonLockState.return_value = {
            "CurrentButtonLockState": "On"
        }
        assert not moco.buttons_enabled
        moco.deviceProperties.GetButtonLockState.return_value = {
            "CurrentButtonLockState": "Off"
        }
        assert moco.buttons_enabled
        moco.deviceProperties.GetButtonLockState.assert_called_with()
        # Setter tests for 'is_visible' property, so this needs to be
        # mocked.
        with mock.patch(
            "pysonos.SoCo.is_visible", new_callable=mock.PropertyMock
        ) as mock_is_visible:
            mock_is_visible.return_value = True
            moco.buttons_enabled = False
            moco.deviceProperties.SetButtonLockState.assert_called_once_with(
                [("DesiredButtonLockState", "On")]
            )
            moco.buttons_enabled = True
            moco.deviceProperties.SetButtonLockState.assert_called_with(
                [("DesiredButtonLockState", "Off")]
            )
            # Check for exception if attempt to set the property on a
            # non-visible speaker.
            mock_is_visible.return_value = False
            with pytest.raises(SoCoNotVisibleException):
                moco.buttons_enabled = True

    def test_soco_set_player_name(self, moco):
        moco.player_name = "μИⅠℂ☺ΔЄ💋"
        moco.deviceProperties.SetZoneAttributes.assert_called_once_with(
            [
                ("DesiredZoneName", "μИⅠℂ☺ΔЄ💋"),
                ("DesiredIcon", ""),
                ("DesiredConfiguration", ""),
            ]
        )

    def test_create_stereo_pair(self, moco):
        """Tests for a well-formed call to create a stereo pair.

        Creates a SoCo object for the slave (RH) speaker, and
        checks for the correct call with the correct parameters.
        """
        moco2 = mock.Mock()
        moco2.uid = "RINCON_000XXY1400"
        moco.create_stereo_pair(moco2)
        moco.deviceProperties.AddBondedZones.assert_called_once_with(
            [("ChannelMapSet", "RINCON_000XXX1400:LF,LF;RINCON_000XXY1400:RF,RF")]
        )

    def test_separate_stereo_pair(self, moco):
        """Tests for a well-formed call to separate a stereo pair."""
        moco.separate_stereo_pair()
        moco.deviceProperties.RemoveBondedZones.assert_called_once_with(
            [("ChannelMapSet", ""), ("KeepGrouped", "0")]
        )

    def test_get_battery_info(self, moco):
        url = "http://" + moco.ip_address + ":1400/status/batterystatus"

        # A speaker that returns battery information
        with requests_mock.Mocker() as m:
            response_text = (
                '<?xml version="1.0" ?>\n<?xml-stylesheet type="text/xsl"'
                + 'href="/xml/review.xsl"?><ZPSupportInfo><LocalBatteryStatus>\n'
                + '<Data name="Health">GREEN</Data>\n<Data name="Level">100</Data>\n'
                + '<Data name="Temperature">NORMAL</Data>\n'
                + '<Data name="PowerSource">SONOS_CHARGING_RING</Data>\n'
                + "</LocalBatteryStatus></ZPSupportInfo>"
            )
            m.get(url, text=response_text)
            assert moco.get_battery_info() == {
                "Health": "GREEN",
                "Level": 100,
                "Temperature": "NORMAL",
                "PowerSource": "SONOS_CHARGING_RING",
            }

        # A speaker that doesn't have battery information
        with requests_mock.Mocker() as m:
            response_text = (
                '<?xml version="1.0" ?>\n'
                + '<?xml-stylesheet type="text/xsl" href="/xml/review.xsl"?>'
                + "<ZPSupportInfo></ZPSupportInfo>"
            )
            m.get(url, text=response_text)
            with pytest.raises(NotSupportedException):
                moco.get_battery_info()

        # A network request that fails
        with requests_mock.Mocker() as m:
            m.get(url, status_code=404)
            with pytest.raises(ConnectionError):
                moco.get_battery_info()


class TestZoneGroupTopology:
    def test_soco_uid(self, moco_zgs):
        assert moco_zgs.uid == "RINCON_000XXX1400"

    def test_soco_is_visible(self, moco_zgs):
        assert moco_zgs.is_visible

    def test_soco_is_bridge(self, moco_zgs):
        assert not moco_zgs.is_bridge

    def test_soco_is_coordinator(self, moco_zgs):
        assert moco_zgs.is_coordinator

    def test_all_groups(self, moco_zgs):
        groups = moco_zgs.all_groups
        assert len(groups) == 3
        # Check 3 unique groups
        assert len(set(groups)) == 3
        for group in groups:
            assert isinstance(group, ZoneGroup)

    def test_all_groups_have_coordinator(self, moco_zgs):
        for group in moco_zgs.all_groups:
            print("GROUP", group.coordinator)
            print(group)
            assert group.coordinator is not None

    def test_group(self, moco_zgs):
        assert isinstance(moco_zgs.group, ZoneGroup)
        assert moco_zgs in moco_zgs.group

    def test_all_zones(selfself, moco_zgs):
        zones = moco_zgs.all_zones
        assert len(zones) == 4
        assert len(set(zones)) == 4
        for zone in zones:
            assert isinstance(zone, SoCo)
        assert moco_zgs in zones

    def test_visible_zones(selfself, moco_zgs):
        zones = moco_zgs.visible_zones
        assert len(zones) == 4
        assert len(set(zones)) == 4
        for zone in zones:
            assert isinstance(zone, SoCo)
        assert moco_zgs in zones

    def test_group_label(selfself, moco_zgs):
        g = moco_zgs.group
        # Have to mock out group members zone group state here since
        # g.members is parsed from the XML.
        for speaker in g.members:
            speaker.zoneGroupTopology.GetZoneGroupState.return_value = {
                "ZoneGroupState": ZGS
            }
        assert g.label == "Kitchen, Living Room"

    def test_group_short_label(selfself, moco_zgs):
        g = moco_zgs.group
        # Have to mock out group members zone group state here since
        # g.members is parsed from the XML.
        for speaker in g.members:
            speaker.zoneGroupTopology.GetZoneGroupState.return_value = {
                "ZoneGroupState": ZGS
            }
        assert g.short_label == "Kitchen + 1"

    def test_group_volume(self, moco_zgs):
        g = moco_zgs.group
        c = moco_zgs.group.coordinator
        c.groupRenderingControl.GetGroupVolume.return_value = {"CurrentVolume": 50}
        assert g.volume == 50
        c.groupRenderingControl.GetGroupVolume.assert_called_once_with(
            [("InstanceID", 0)]
        )
        g.volume = 75
        c.groupRenderingControl.SetGroupVolume.assert_called_once_with(
            [("InstanceID", 0), ("DesiredVolume", 75)]
        )

    def test_group_mute(self, moco_zgs):
        g = moco_zgs.group
        c = moco_zgs.group.coordinator
        c.groupRenderingControl.GetGroupMute.return_value = {"CurrentMute": "0"}
        assert g.mute is False
        c.groupRenderingControl.GetGroupMute.assert_called_once_with(
            [("InstanceID", 0)]
        )
        g.mute = True
        c.groupRenderingControl.SetGroupMute.assert_called_once_with(
            [("InstanceID", 0), ("DesiredMute", "1")]
        )

    def test_set_relative_group_volume(self, moco_zgs):
        g = moco_zgs.group
        c = moco_zgs.group.coordinator
        c.groupRenderingControl.SetRelativeGroupVolume.return_value = {
            "NewVolume": "75"
        }
        new_volume = g.set_relative_volume(25)
        c.groupRenderingControl.SetRelativeGroupVolume.assert_called_once_with(
            [("InstanceID", 0), ("Adjustment", 25)]
        )
        assert new_volume == 75


def test_only_on_master_true(moco_only_on_master):
    with mock.patch(
        "pysonos.SoCo.is_coordinator", new_callable=mock.PropertyMock
    ) as is_coord:
        is_coord.return_value = True
        moco_only_on_master.play()
        is_coord.assert_called_once_with()


def test_not_on_master_false(moco_only_on_master):
    with mock.patch(
        "pysonos.SoCo.is_coordinator", new_callable=mock.PropertyMock
    ) as is_coord:
        is_coord.return_value = False
        with pytest.raises(SoCoSlaveException):
            moco_only_on_master.play()
        is_coord.assert_called_once_with()



"""Tests for the xml module."""


from pysonos import xml


def test_ns_tag():
    """Test the ns_tag function."""
    namespaces = [
        "http://purl.org/dc/elements/1.1/",
        "urn:schemas-upnp-org:metadata-1-0/upnp/",
        "urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/",
    ]
    for ns_in, namespace in zip(["dc", "upnp", ""], namespaces):
        res = xml.ns_tag(ns_in, "testtag")
        correct = "{{{}}}{}".format(namespace, "testtag")
        assert res == correct



#! /usr/bin/env python


""" Dump all known UPnP commands """

import argparse
import re

import soco
import soco.services


def main():
    """ Run the main script """
    parser = argparse.ArgumentParser(
        prog='',
        description='Dump data about Sonos services'
    )
    parser.add_argument(
        '-d', '--device',
        default=None,
        help="The ip address of the device to query. "
             "If none is supplied, a random device will be used"
    )
    parser.add_argument(
        '-s', '--service',
        default=None,
        help="Dump data relating to services matching this regexp "
             "only, e.g. %(prog)s -s GroupRenderingControl"
    )

    args = parser.parse_args()

    # get a zone player - any one will do
    if args.device:
        device = soco.SoCo(args.device)
    else:
        device = soco.discovery.any_soco()
    print("Querying %s" % device.player_name)
    # loop over each of the available services
    # pylint: disable=no-member
    services = (srv(device) for srv in soco.services.Service.__subclasses__())

    for srv in services:
        if args.service is None or re.search(
                args.service, srv.service_type):
            print_details(srv)


def print_details(srv):
    """ Print the details of a service
    """
    name = srv.service_type
    box = "=" * 79
    print("{0}\n|{1:^77}|\n{0}\n".format(box, name))
    for action in srv.iter_actions():
        print(action.name)
        print("~" * len(action.name))
        print("\n  Input")
        for arg in action.in_args:
            print("    ", arg)
        print("\n  Output")
        for arg in action.out_args:
            print("    ", arg)

        print("\n\n")


if __name__ == '__main__':
    main()



"""Tests for the SoCoSingletonBase and _ArgsSingleton classes in core."""


import pytest

from pysonos.core import _SocoSingletonBase as Base


class ASingleton(Base):
    def __init__(self, arg):
        pass


class AnotherSingleton(ASingleton):
    pass


class ThirdSingleton(Base):
    _class_group = "somegroup"

    def __init__(self, arg):
        pass


class FourthSingleton(ASingleton):
    _class_group = "somegroup"
    pass


def test_singleton():
    """Check basic functionality.

    For a given arg, there is only one instance
    """
    assert ASingleton("aa") is ASingleton("aa")
    assert ASingleton("aa") is not ASingleton("bb")


def test_singleton_inherit():
    """Check that subclasses behave properly."""
    assert ASingleton("aa") is not AnotherSingleton("aa")
    assert AnotherSingleton("aa") is AnotherSingleton("aa")


def test_class_group_singleton():
    """Check _class_group functionality.

    For a given arg, instances of FourthGroup are Instances of
    ThirdGroup because they share a `_class_group` valur
    """
    assert ThirdSingleton("aa") is FourthSingleton("aa")
    assert ThirdSingleton("aa") is not FourthSingleton("bb")
    assert ThirdSingleton("aa") is not ASingleton("aa")



# pylint: disable=not-context-manager

# NOTE: The pylint not-content-manager warning is disabled pending the fix of
# a bug in pylint. See https://github.com/PyCQA/pylint/issues/782


"""Base classes used by :py:mod:`soco.events` and
:py:mod:`soco.events_twisted`."""


import atexit
import logging
import socket
import time
import threading
import weakref
from queue import Queue

from . import config
from .data_structures_entry import from_didl_string
from .exceptions import SoCoException, SoCoFault, EventParseException
from .utils import camel_to_underscore
from .xml import XML

log = logging.getLogger(__name__)  # pylint: disable=C0103


# pylint: disable=too-many-branches
def parse_event_xml(xml_event):
    """Parse the body of a UPnP event.

    Args:
        xml_event (bytes): bytes containing the body of the event encoded
            with utf-8.

    Returns:
        dict: A dict with keys representing the evented variables. The
        relevant value will usually be a string representation of the
        variable's value, but may on occasion be:

        * a dict (eg when the volume changes, the value will itself be a
          dict containing the volume for each channel:
          :code:`{'Volume': {'LF': '100', 'RF': '100', 'Master': '36'}}`)
        * an instance of a `DidlObject` subclass (eg if it represents
          track metadata).
        * a `SoCoFault` (if a variable contains illegal metadata)
    """

    result = {}
    tree = XML.fromstring(xml_event)
    # property values are just under the propertyset, which
    # uses this namespace
    properties = tree.findall("{urn:schemas-upnp-org:event-1-0}property")
    for prop in properties:  # pylint: disable=too-many-nested-blocks
        for variable in prop:
            # Special handling for a LastChange event specially. For details on
            # LastChange events, see
            # http://upnp.org/specs/av/UPnP-av-RenderingControl-v1-Service.pdf
            # and http://upnp.org/specs/av/UPnP-av-AVTransport-v1-Service.pdf
            if variable.tag == "LastChange":
                last_change_tree = XML.fromstring(variable.text.encode("utf-8"))
                # We assume there is only one InstanceID tag. This is true for
                # Sonos, as far as we know.
                # InstanceID can be in one of two namespaces, depending on
                # whether we are looking at an avTransport event, a
                # renderingControl event, or a Queue event
                # (there, it is named QueueID)
                instance = last_change_tree.find(
                    "{urn:schemas-upnp-org:metadata-1-0/AVT/}InstanceID"
                )
                if instance is None:
                    instance = last_change_tree.find(
                        "{urn:schemas-upnp-org:metadata-1-0/RCS/}InstanceID"
                    )
                if instance is None:
                    instance = last_change_tree.find(
                        "{urn:schemas-sonos-com:metadata-1-0/Queue/}QueueID"
                    )
                # Look at each variable within the LastChange event
                for last_change_var in instance:
                    tag = last_change_var.tag
                    # Remove any namespaces from the tags
                    if tag.startswith("{"):
                        tag = tag.split("}", 1)[1]
                    # Un-camel case it
                    tag = camel_to_underscore(tag)
                    # Now extract the relevant value for the variable.
                    # The UPnP specs suggest that the value of any variable
                    # evented via a LastChange Event will be in the 'val'
                    # attribute, but audio related variables may also have a
                    # 'channel' attribute. In addition, it seems that Sonos
                    # sometimes uses a text value instead: see
                    # http://forums.sonos.com/showthread.php?t=34663
                    value = last_change_var.get("val")
                    if value is None:
                        value = last_change_var.text
                    # If DIDL metadata is returned, convert it to a music
                    # library data structure
                    if value.startswith("<DIDL-Lite"):
                        # Wrap any parsing exception in a SoCoFault, so the
                        # user can handle it
                        try:
                            value = from_didl_string(value)[0]
                        except SoCoException as original_exception:
                            log.debug(
                                "Event contains illegal metadata"
                                "for '%s'.\n"
                                "Error message: '%s'\n"
                                "The result will be a SoCoFault.",
                                tag,
                                str(original_exception),
                            )
                            event_parse_exception = EventParseException(
                                tag, value, original_exception
                            )
                            value = SoCoFault(event_parse_exception)
                    channel = last_change_var.get("channel")
                    if channel is not None:
                        if result.get(tag) is None:
                            result[tag] = {}
                        result[tag][channel] = value
                    else:
                        result[tag] = value
            else:
                result[camel_to_underscore(variable.tag)] = variable.text
    return result


class Event:
    """A read-only object representing a received event.

    The values of the evented variables can be accessed via the ``variables``
    dict, or as attributes on the instance itself. You should treat all
    attributes as read-only.

    Args:
        sid (str): the subscription id.
        seq (str): the event sequence number for that subscription.
        timestamp (str): the time that the event was received (from Python's
            `time.time` function).
        service (str): the service which is subscribed to the event.
        variables (dict, optional): contains the ``{names: values}`` of the
            evented variables. Defaults to `None`. The values may be
            `SoCoFault` objects if the metadata could not be parsed.

    Raises:
        AttributeError:  Not all attributes are returned with each event. An
            `AttributeError` will be raised if you attempt to access as an
            attribute a variable which was not returned in the event.

    Example:

        >>> print event.variables['transport_state']
        'STOPPED'
        >>> print event.transport_state
        'STOPPED'

    """

    # pylint: disable=too-few-public-methods, too-many-arguments

    def __init__(self, sid, seq, service, timestamp, variables=None):
        # Initialisation has to be done like this, because __setattr__ is
        # overridden, and will not allow direct setting of attributes
        self.__dict__["sid"] = sid
        self.__dict__["seq"] = seq
        self.__dict__["timestamp"] = timestamp
        self.__dict__["service"] = service
        self.__dict__["variables"] = variables if variables is not None else {}

    def __getattr__(self, name):
        if name in self.variables:
            return self.variables[name]
        else:
            raise AttributeError("No such attribute: %s" % name)

    def __setattr__(self, name, value):
        """Disable (most) attempts to set attributes.

        This is not completely foolproof. It just acts as a warning! See
        `object.__setattr__`.
        """
        raise TypeError("Event object does not support attribute assignment")


class EventNotifyHandlerBase:
    """Base class for `soco.events.EventNotifyHandler` and
    `soco.events_twisted.EventNotifyHandler`.
    """

    # pylint: disable=too-many-public-methods

    def handle_notification(self, headers, content):
        """Handle a ``NOTIFY`` request by building an `Event` object and
        sending it to the relevant Subscription object.

        A ``NOTIFY`` request will be sent by a Sonos device when a state
        variable changes. See the `UPnP Spec §4.3 [pdf]
        <http://upnp.org/specs/arch/UPnP-arch
        -DeviceArchitecture-v1.1.pdf>`_  for details.

        Args:
            headers (dict): A dict of received headers.
            content (str): A string of received content.
        Note:
            Each of the :py:mod:`soco.events` and the
            :py:mod:`soco.events_twisted` modules has a **subscriptions_map**
            object which keeps a record of Subscription objects. The
            *get_subscription* method of the **subscriptions_map** object is
            used to look up the subscription to which the event relates. When
            the Event Listener runs in a thread (the default), a lock is
            used by this method for thread safety. The *send_event*
            method of the relevant Subscription will first check to see
            whether the *callback* variable of the Subscription has been
            set. If it has been and is callable, then the *callback*
            will be called with the `Event` object. Otherwise, the `Event`
            object will be sent to the event queue of the Subscription
            object. The *callback* variable of the Subscription object is
            intended for use only if :py:mod:`soco.events_twisted` is being
            used, as calls to it are not threadsafe.

            This method calls the log_event method, which must be overridden
            in the class that inherits from this class.
        """

        timestamp = time.time()
        seq = headers["seq"]  # Event sequence number
        sid = headers["sid"]  # Event Subscription Identifier
        # find the relevant service from the sid
        # pylint: disable=no-member
        subscription = self.subscriptions_map.get_subscription(sid)
        # It might have been removed by another thread
        if subscription:
            service = subscription.service
            self.log_event(seq, service.service_id, timestamp)
            log.debug("Event content: %s", content)
            variables = parse_event_xml(content)
            # Build the Event object
            event = Event(sid, seq, service, timestamp, variables)
            # pass the event details on to the service so it can update
            # its cache.
            # pylint: disable=protected-access
            service._update_cache_on_event(event)
            # Pass the event on for handling
            subscription.send_event(event)
        else:
            log.info("No service registered for %s", sid)

    # pylint: disable=missing-docstring
    def log_event(self, seq, service_id, timestamp):
        raise NotImplementedError


class EventListenerBase:
    """Base class for `soco.events.EventListener` and
    `soco.events_twisted.EventListener`.
    """

    def __init__(self):
        #: `bool`: Indicates whether the server is currently running
        self.is_running = False
        self._start_lock = threading.Lock()
        #: `tuple`: The address (ip, port) on which the server is
        #: configured to listen.
        # Empty for the moment. (It is set in `start`)
        self.address = ()
        #: `int`: Port on which to listen.
        self.requested_port_number = config.EVENT_LISTENER_PORT

    def start(self, any_zone):
        """Start the event listener listening on the local machine.

        Args:
            any_zone (SoCo): Any Sonos device on the network. It does not
                matter which device. It is used only to find a local IP
                address reachable by the Sonos net.

        """

        # Find our local network IP address which is accessible to the
        # Sonos net, see http://stackoverflow.com/q/166506
        with self._start_lock:
            if self.is_running:
                return
            # Use configured IP address if there is one, else detect
            # automatically.
            ip_address = get_listen_ip(any_zone.ip_address)
            if not ip_address:
                log.exception("Could not start Event Listener: check network.")
                # Otherwise, no point trying to start server
                return
            port = self.listen(ip_address)
            if not port:
                return
            self.address = (ip_address, port)
            self.is_running = True
            log.debug("Event Listener started")

    def stop(self):
        """Stop the Event Listener."""
        if not self.is_running:
            return
        self.is_running = False
        self.stop_listening(self.address)
        log.debug("Event Listener stopped")

    # pylint: disable=missing-docstring
    def listen(self, ip_address):
        """Start the event listener listening on the local machine.
        This method is called by `start`.

        Args:
            ip_address (str): The local network interface on which the server
                should start listening.
        Returns:
            int: The port on which the server is listening.

        Note:
            This method must be overridden in the class that inherits from
            this class.
        """
        raise NotImplementedError

    # pylint: disable=missing-docstring
    def stop_listening(self, address):
        """Stop the listener.

        Note:
            This method must be overridden in the class that inherits from
            this class.
        """
        raise NotImplementedError


class SubscriptionBase:
    """Base class for `soco.events.Subscription` and
    `soco.events_twisted.Subscription`
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, service, event_queue=None):
        """
        Args:
            service (Service): The SoCo `Service` to which the subscription
                 should be made.
            event_queue (:class:`~queue.Queue`): A queue on which received
                events will be put. If not specified, a queue will be
                created and used.
        """
        self.service = service
        #: `str`: A unique ID for this subscription
        self.sid = None
        #: `int`: The amount of time in seconds until the subscription expires.
        self.timeout = None
        #: `bool`: An indication of whether the subscription is subscribed.
        self.is_subscribed = False
        #: :class:`~queue.Queue`: The queue on which events are placed.
        self.events = Queue() if event_queue is None else event_queue
        #: `int`: The period (seconds) for which the subscription is requested
        self.requested_timeout = None
        #: :py:obj:`function`: an optional function to be called if an
        #: exception occurs upon autorenewal. This will be called with the
        #: exception (or failure, when using :py:mod:`soco.events_twisted`)
        #: as its only parameter. This function must be threadsafe (unless
        #: :py:mod:`soco.events_twisted` is being used).
        self.auto_renew_fail = None
        # A flag to make sure that an unsubscribed instance is not
        # resubscribed
        self._has_been_unsubscribed = False
        # The time when the subscription was made
        self._timestamp = None

    def subscribe(self, requested_timeout=None, auto_renew=False):
        """Subscribe to the service.

        If requested_timeout is provided, a subscription valid for that number
        of seconds will be requested, but not guaranteed. Check
        `timeout` on return to find out what period of validity is
        actually allocated.

        Note:
            SoCo will try to unsubscribe any subscriptions which are still
            subscribed on program termination, but it is good practice for
            you to clean up by making sure that you call :meth:`unsubscribe`
            yourself.

        Args:
            requested_timeout(int, optional): The timeout to be requested.
            auto_renew (bool, optional): If `True`, renew the subscription
                automatically shortly before timeout. Default `False`.

        """

        # TIMEOUT is provided for in the UPnP spec, but it is not clear if
        # Sonos pays any attention to it. A timeout of 86400 secs always seems
        # to be allocated
        self.requested_timeout = requested_timeout
        if self.is_subscribed:
            raise SoCoException(
                "Cannot subscribe Subscription instance more than once. "
                + "Use renew instead"
            )
        if self._has_been_unsubscribed:
            raise SoCoException(
                "Cannot resubscribe Subscription instance once unsubscribed"
            )
        service = self.service
        # The Event Listener must be running, so start it if not
        # pylint: disable=no-member
        if not self.event_listener.is_running:
            self.event_listener.start(service.soco)
        # an event subscription looks like this:
        # SUBSCRIBE publisher path HTTP/1.1
        # HOST: publisher host:publisher port
        # CALLBACK: <delivery URL>
        # NT: upnp:event
        # TIMEOUT: Second-requested subscription duration (optional)

        # pylint: disable=unbalanced-tuple-unpacking
        ip_address, port = self.event_listener.address

        if config.EVENT_ADVERTISE_IP:
            ip_address = config.EVENT_ADVERTISE_IP

        headers = {
            "Callback": "<http://{}:{}>".format(ip_address, port),
            "NT": "upnp:event",
        }
        if requested_timeout is not None:
            headers["TIMEOUT"] = "Second-{}".format(requested_timeout)

        # pylint: disable=missing-docstring
        def success(headers):
            self.sid = headers["sid"]
            timeout = headers["timeout"]
            # According to the spec, timeout can be "infinite" or "second-123"
            # where 123 is a number of seconds.  Sonos uses "Second-123"
            # (with a capital letter)
            if timeout.lower() == "infinite":
                self.timeout = None
            else:
                self.timeout = int(timeout.lstrip("Second-"))
            self._timestamp = time.time()
            self.is_subscribed = True
            log.debug(
                "Subscribed to %s, sid: %s",
                service.base_url + service.event_subscription_url,
                self.sid,
            )
            # Register the subscription so it can be looked up by sid
            # and unsubscribed at exit
            self.subscriptions_map.register(self)

            # Set up auto_renew
            if not auto_renew:
                return
            # Autorenew just before expiry, say at 85% of self.timeout seconds
            interval = self.timeout * 85 / 100
            self._auto_renew_start(interval)

        # Lock out EventNotifyHandler during registration.
        # If events_twisted is used, this lock should always be
        # available, since threading is not being used. This is to prevent
        # the EventNotifyHandler from sending a notification before the
        # subscription has been registered.
        with self.subscriptions_map.subscriptions_lock:
            return self._request(
                "SUBSCRIBE",
                service.base_url + service.event_subscription_url,
                headers,
                success,
            )

    def renew(self, requested_timeout=None, is_autorenew=False):
        """renew(requested_timeout=None)
        Renew the event subscription.
        You should not try to renew a subscription which has been
        unsubscribed, or once it has expired.

        Args:
            requested_timeout (int, optional): The period for which a renewal
                request should be made. If None (the default), use the timeout
                requested on subscription.
            is_autorenew (bool, optional): Whether this is an autorenewal.

        """
        # NB This code may be called from a separate thread when
        # subscriptions are auto-renewed. Be careful to ensure thread-safety

        if is_autorenew:
            log_msg = "Autorenewing subscription %s"
        else:
            log_msg = "Renewing subscription %s"
        log.debug(log_msg, self.sid)

        if self._has_been_unsubscribed:
            raise SoCoException("Cannot renew subscription once unsubscribed")
        if not self.is_subscribed:
            raise SoCoException("Cannot renew subscription before subscribing")
        if self.time_left == 0:
            raise SoCoException("Cannot renew subscription after expiry")

        # SUBSCRIBE publisher path HTTP/1.1
        # HOST: publisher host:publisher port
        # SID: uuid:subscription UUID
        # TIMEOUT: Second-requested subscription duration (optional)
        headers = {"SID": self.sid}
        if requested_timeout is None:
            requested_timeout = self.requested_timeout
        if requested_timeout is not None:
            headers["TIMEOUT"] = "Second-{}".format(requested_timeout)

        # pylint: disable=missing-docstring
        def success(headers):
            timeout = headers["timeout"]
            # According to the spec, timeout can be "infinite" or "second-123"
            # where 123 is a number of seconds.  Sonos uses "Second-123"
            # (with a capital letter)
            if timeout.lower() == "infinite":
                self.timeout = None
            else:
                self.timeout = int(timeout.lstrip("Second-"))
            self._timestamp = time.time()
            self.is_subscribed = True
            log.debug(
                "Renewed subscription to %s, sid: %s",
                self.service.base_url + self.service.event_subscription_url,
                self.sid,
            )

        return self._request(
            "SUBSCRIBE",
            self.service.base_url + self.service.event_subscription_url,
            headers,
            success,
        )

    def unsubscribe(self):
        """unsubscribe()
        Unsubscribe from the service's events.
        Once unsubscribed, a Subscription instance should not be reused
        """
        # Trying to unsubscribe if already unsubscribed, or not yet
        # subscribed, fails silently
        if self._has_been_unsubscribed or not self.is_subscribed:
            return None

        # If the subscription has timed out, an attempt to
        # unsubscribe from it will fail silently.
        if self.time_left == 0:
            return None

        # Send an unsubscribe request like this:
        # UNSUBSCRIBE publisher path HTTP/1.1
        # HOST: publisher host:publisher port
        # SID: uuid:subscription UUID
        headers = {"SID": self.sid}

        # pylint: disable=missing-docstring, unused-argument
        def success(*arg):
            log.debug(
                "Unsubscribed from %s, sid: %s",
                self.service.base_url + self.service.event_subscription_url,
                self.sid,
            )

        return self._request(
            "UNSUBSCRIBE",
            self.service.base_url + self.service.event_subscription_url,
            headers,
            success,
            self._cancel_subscription,
        )

    def send_event(self, event):
        """Send an `Event` to self.callback or self.events.
        If self.callback is set and is callable, it will be called with the
        `Event` as the only parameter. Otherwise the `Event` will be sent to
        self.events. As self.callback is not threadsafe, it should be set
        only if `soco.events_twisted.Subscription` is being used.

        Args:
            event(Event): The `Event` to send to self.callback or
                self.events.

        """
        if hasattr(self, "callback"):
            # pylint: disable=no-member
            callback = self.callback
        else:
            callback = None
        if callback and hasattr(callback, "__call__"):
            callback(event)
        else:
            try:
                self.events.put(event)
            # pylint: disable=broad-except
            except Exception as ex:
                log.warning("Error putting event %s, ex=%s", event, ex)

    # pylint: disable=missing-docstring
    def _auto_renew_start(self, interval):
        """Starts the auto_renew thread.

        Note:
            This method must be overridden in the class that inherits from
            this class.
        """
        raise NotImplementedError

    # pylint: disable=missing-docstring
    def _auto_renew_cancel(self):
        """Cancels the auto_renew thread.

        Note:
            This method must be overridden in the class that inherits from
            this class.
        """
        raise NotImplementedError

    # pylint: disable=missing-docstring, too-many-arguments
    def _request(self, method, url, headers, success, unconditional=None):
        """Send a HTTP request

        Args:
            method (str): 'SUBSCRIBE' or 'UNSUBSCRIBE'.
            url (str): The full endpoint to which the request is being sent.
            headers (dict): A dict of headers, each key and each value being
                of type `str`.
            success (function): A function to be called if the
                request succeeds. The function will be called with a dict
                of response headers as its only parameter.
            unconditional (function): An optional function to be called after
                the request is complete, regardless of its success. Takes
                no parameters.

        Note:
            This method must be overridden in the class that inherits from
            this class.
        """
        raise NotImplementedError

    # pylint: disable=missing-docstring
    def _cancel_subscription(self, msg=None):
        # unregister subscription
        # pylint: disable=no-member
        self.subscriptions_map.unregister(self)
        # Stop the event listener, if there are no other subscriptions
        if self.subscriptions_map.count == 0:
            self.event_listener.stop()
        # No need to do any more if this flag has been set to True
        if self._has_been_unsubscribed:
            return
        self.is_subscribed = False
        # Set the self._has_been_unsubscribed flag now
        # to prevent reuse of the subscription, even if
        # an attempt to unsubscribe fails
        self._has_been_unsubscribed = True
        self._timestamp = None
        # Cancel any auto renew
        self._auto_renew_cancel()
        if msg:
            log.info(msg)

    @property
    def time_left(self):
        """
        `int`: The amount of time left until the subscription expires (seconds)
        If the subscription is unsubscribed (or not yet subscribed),
        `time_left` is 0.
        """
        if self._timestamp is None:
            return 0
        else:
            time_left = self.timeout - (time.time() - self._timestamp)
            return time_left if time_left > 0 else 0

    def __enter__(self):
        if not self.is_subscribed:
            self.subscribe()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unsubscribe()


class SubscriptionsMap:
    """Maintains a mapping of sids to `soco.events.Subscription` instances
    and the thread safe lock to go with it. Registers each subscription to
    be unsubscribed at exit.

    `SubscriptionsMapTwisted` inherits from this class.

    """

    def __init__(self):
        super().__init__()
        #: `weakref.WeakValueDictionary`: Thread safe mapping.
        #: Used to store a mapping of sid to subscription
        self.subscriptions = weakref.WeakValueDictionary()
        # The lock to go with it
        # You must only ever access the mapping in the context of this lock,
        # eg:
        #   with self.subscriptions_lock:
        #       queue = self.subscriptions[sid].events
        #: `threading.Lock`: for use with `subscriptions`
        self.subscriptions_lock = threading.Lock()

    def register(self, subscription):
        """Register a subscription by updating local mapping of sid to
        subscription and registering it to be unsubscribed at exit.

        Args:
            subscription(`soco.events.Subscription`): the subscription
                to be registered.

        """
        # Add the queue to the master dict of subscriptions so it can be
        # looked up by sid. The subscriptions_lock is not used here as
        # it is used in Subscription.subscribe() in the events_base
        # module, from which the register function is called.
        self.subscriptions[subscription.sid] = subscription
        # Register subscription to be unsubscribed at exit if still alive
        # This will not happen if exit is abnormal (eg in response to a
        # signal or fatal interpreter error - see the docs for `atexit`).
        atexit.register(subscription.unsubscribe)

    def unregister(self, subscription):
        """Unregister a subscription by updating local mapping of sid to
        subscription instances.

        Args:
            subscription(`soco.events.Subscription`): the subscription
                to be unregistered.

        When using :py:mod:`soco.events_twisted`, an instance of
        `soco.events_twisted.Subscription` will be unregistered.

        """
        with self.subscriptions_lock:
            try:
                del self.subscriptions[subscription.sid]
            except KeyError:
                pass

    def get_subscription(self, sid):
        """Look up a subscription from a sid.

            Args:
                sid(str): The sid from which to look up the subscription.

            Returns:
                `soco.events.Subscription`: The subscription relating
                to that sid.

        When using :py:mod:`soco.events_twisted`, an instance of
        `soco.events_twisted.Subscription` will be returned.

        """
        with self.subscriptions_lock:
            return self.subscriptions.get(sid)

    @property
    def count(self):
        """
        `int`: The number of active subscriptions.
        """
        with self.subscriptions_lock:
            return len(self.subscriptions)


def get_listen_ip(ip_address):
    """Find the listen ip address."""
    if config.EVENT_LISTENER_IP:
        return config.EVENT_LISTENER_IP
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        sock.connect((ip_address, config.EVENT_LISTENER_PORT))
        return sock.getsockname()[0]
    except socket.error:
        return None
    finally:
        sock.close()



import time
import hashlib
import json

import requests
from flask import Flask, render_template, url_for

from soco import SoCo

app = Flask(__name__)

app.config.from_pyfile("settings.py")

sonos = SoCo(app.config["SPEAKER_IP"])


def gen_sig():
    return hashlib.md5(
        (
            app.config["ROVI_API_KEY"]
            + app.config["ROVI_SHARED_SECRET"]
            + repr(int(time.time()))
        ).encode("utf-8")
    ).hexdigest()


def get_track_image(artist, album):
    blank_image = url_for("static", filename="img/blank.jpg")
    if "ROVI_SHARED_SECRET" not in app.config:
        return blank_image
    elif "ROVI_API_KEY" not in app.config:
        return blank_image

    headers = {"Accept-Encoding": "gzip"}
    req = requests.get(
        "http://api.rovicorp.com/recognition/v2.1/music/match/album?apikey="
        + app.config["ROVI_API_KEY"]
        + "&sig="
        + gen_sig()
        + "&name= "
        + album
        + "&performername="
        + artist
        + "&include=images&size=1",
        headers=headers,
    )

    if req.status_code != requests.codes.ok:
        return blank_image

    result = json.loads(req.content)
    try:
        return result["matchResponse"]["results"][0]["album"]["images"][0]["front"][3][
            "url"
        ]
    except (KeyError, IndexError):
        return blank_image


@app.route("/play")
def play():
    sonos.play()
    return "Ok"


@app.route("/pause")
def pause():
    sonos.pause()
    return "Ok"


@app.route("/next")
def next():
    sonos.next()
    return "Ok"


@app.route("/previous")
def previous():
    sonos.previous()
    return "Ok"


@app.route("/info-light")
def info_light():
    track = sonos.get_current_track_info()
    return json.dumps(track)


@app.route("/info")
def info():
    track = sonos.get_current_track_info()
    track["image"] = get_track_image(track["artist"], track["album"])
    return json.dumps(track)


@app.route("/")
def index():
    track = sonos.get_current_track_info()
    track["image"] = get_track_image(track["artist"], track["album"])
    return render_template("index.html", track=track)


if __name__ == "__main__":
    app.run(debug=True)



#!/usr/bin/env python

import io
import re
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    # Code from here: https://pytest.org/latest/goodpractises.html

    def finalize_options(self):
        TestCommand.finalize_options(self)
        # We don't run integration tests which need an actual Sonos device
        self.test_args = ["-m", "not integration"]
        self.test_suite = True

    def run_tests(self):
        # Import here, because outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.test_args)
        sys.exit(errno)


src = io.open("pysonos/__init__.py", encoding="utf-8").read()
metadata = dict(re.findall('__([a-z]+)__ = "([^"]+)"', src))
docstrings = re.findall('"""(.*?)"""', src, re.MULTILINE | re.DOTALL)

NAME = "pysonos"

PACKAGES = (
    "pysonos",
    "pysonos.plugins",
    "pysonos.music_services",
)

TEST_REQUIREMENTS = list(open("requirements-dev.txt"))
AUTHOR_EMAIL = metadata["author"]
VERSION = metadata["version"]
WEBSITE = metadata["website"]
LICENSE = metadata["license"]
DESCRIPTION = docstrings[0]

CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Topic :: Home Automation",
    "Topic :: Multimedia :: Sound/Audio",
    "Topic :: Multimedia :: Sound/Audio :: Players",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

PYTHON_REQUIRES = ">=3.7"

with open("README.rst", encoding="utf-8") as file:
    LONG_DESCRIPTION = file.read()

# Extract name and e-mail ("Firstname Lastname <mail@example.org>")
AUTHOR, EMAIL = re.match(r"(.*) <(.*)>", AUTHOR_EMAIL).groups()

REQUIREMENTS = list(open("requirements.txt"))

# See https://github.com/SoCo/SoCo/issues/819
EXTRAS_REQUIRE = {"events_asyncio": ["aiohttp"]}

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=EMAIL,
    license=LICENSE,
    url=WEBSITE,
    packages=PACKAGES,
    install_requires=REQUIREMENTS,
    extras_require=EXTRAS_REQUIRE,
    tests_require=TEST_REQUIREMENTS,
    long_description=LONG_DESCRIPTION,
    cmdclass={"test": PyTest},
    classifiers=CLASSIFIERS,
    python_requires=PYTHON_REQUIRES,
)



"""Example implementation of a plugin."""


from ..plugins import SoCoPlugin

__all__ = ["ExamplePlugin"]


class ExamplePlugin(SoCoPlugin):

    """This file serves as an example of a SoCo plugin."""

    def __init__(self, soco, username):
        """Initialize the plugin.

        The plugin can accept any arguments it requires. It should at
        least accept a soco instance which it passes on to the base
        class when calling super's __init__.
        """
        super().__init__(soco)
        self.username = username

    @property
    def name(self):
        return "Example Plugin for {}".format(self.username)

    def music_plugin_play(self):
        """Play some music.

        This is just a reimplementation of the ordinary play function,
        to show how we can use the general upnp methods from soco
        """

        print("Hi,", self.username)

        self.soco.avTransport.Play([("InstanceID", 0), ("Speed", 1)])

    def music_plugin_stop(self):
        """Stop the music.

        This methods shows how, if we need it, we can use the soco
        functionality from inside the plugins
        """

        print("Bye,", self.username)
        self.soco.stop()



#!/usr/bin/env python

# This illustrates how to use SoCo plugins
# an example plugin is provided in soco.plugins.example.ExamplePlugin

import time

from soco import SoCo
from soco.plugins import SoCoPlugin


def main():
    speakers = [speaker.ip_address for speaker in SoCo.discover()]

    if not speakers:
        print("no speakers found, exiting.")
        return

    soco = SoCo(speakers[0])

    # get a plugin by name (eg from a config file)
    myplugin = SoCoPlugin.from_name(
        "soco.plugins.example.ExamplePlugin", soco, "some user"
    )

    # do something with your plugin
    print("Testing", myplugin.name)
    myplugin.music_plugin_play()

    time.sleep(5)

    # create a plugin by normal instantiation
    from soco.plugins.example import ExamplePlugin

    # create a new plugin, pass the soco instance to it
    myplugin = ExamplePlugin(soco, "a user")

    print("Testing", myplugin.name)

    # do something with your plugin
    myplugin.music_plugin_stop()


if __name__ == "__main__":
    main()



# pylint: disable=invalid-name,wrong-import-position,redefined-builtin

"""This class contains XML related utility functions."""


import sys
import re

import xml.etree.ElementTree as XML


# Create regular expression for filtering invalid characters, from:
# http://stackoverflow.com/questions/1707890/
# fast-way-to-filter-illegal-xml-unicode-chars-in-python

illegal_unichrs = [
    (0x00, 0x08),
    (0x0B, 0x0C),
    (0x0E, 0x1F),
    (0x7F, 0x84),
    (0x86, 0x9F),
    (0xD800, 0xDFFF),
    (0xFDD0, 0xFDDF),
    (0xFFFE, 0xFFFF),
    (0x1FFFE, 0x1FFFF),
    (0x2FFFE, 0x2FFFF),
    (0x3FFFE, 0x3FFFF),
    (0x4FFFE, 0x4FFFF),
    (0x5FFFE, 0x5FFFF),
    (0x6FFFE, 0x6FFFF),
    (0x7FFFE, 0x7FFFF),
    (0x8FFFE, 0x8FFFF),
    (0x9FFFE, 0x9FFFF),
    (0xAFFFE, 0xAFFFF),
    (0xBFFFE, 0xBFFFF),
    (0xCFFFE, 0xCFFFF),
    (0xDFFFE, 0xDFFFF),
    (0xEFFFE, 0xEFFFF),
    (0xFFFFE, 0xFFFFF),
    (0x10FFFE, 0x10FFFF),
]

illegal_ranges = [
    "{}-{}".format(chr(low), chr(high))
    for (low, high) in illegal_unichrs
    if low < sys.maxunicode
]

illegal_xml_re = re.compile("[%s]" % "".join(illegal_ranges))


#: Commonly used namespaces, and abbreviations, used by `ns_tag`.
NAMESPACES = {
    "dc": "http://purl.org/dc/elements/1.1/",
    "upnp": "urn:schemas-upnp-org:metadata-1-0/upnp/",
    "": "urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/",
    "ms": "http://www.sonos.com/Services/1.1",
    "r": "urn:schemas-rinconnetworks-com:metadata-1-0/",
}

# Register common namespaces to assist in serialisation (avoids the ns:0
# prefixes in XML output )
for prefix, uri in NAMESPACES.items():
    XML.register_namespace(prefix, uri)


def ns_tag(ns_id, tag):
    """Return a namespace/tag item.

    Args:
        ns_id (str): A namespace id, eg ``"dc"`` (see `NAMESPACES`)
        tag (str): An XML tag, eg ``"author"``

    Returns:
        str: A fully qualified tag.

    The ns_id is translated to a full name space via the :const:`NAMESPACES`
    constant::

        >>> xml.ns_tag('dc','author')
        '{http://purl.org/dc/elements/1.1/}author'
    """
    return "{{{}}}{}".format(NAMESPACES[ns_id], tag)



# Disable while we have Python 2.x compatability
# pylint: disable=useless-object-inheritance

"""Access to the Music Library.

The Music Library is the collection of music stored on your local network.
For access to third party music streaming services, see the
`music_service` module."""


import logging

from urllib.parse import quote as quote_url

import xmltodict

from . import discovery
from .data_structures import SearchResult, DidlResource, DidlObject, DidlMusicAlbum
from .data_structures_entry import from_didl_string
from .exceptions import SoCoUPnPException
from .utils import url_escape_path, really_unicode, camel_to_underscore

_LOG = logging.getLogger(__name__)


class MusicLibrary:
    """The Music Library."""

    # Key words used when performing searches
    SEARCH_TRANSLATION = {
        "artists": "A:ARTIST",
        "album_artists": "A:ALBUMARTIST",
        "albums": "A:ALBUM",
        "genres": "A:GENRE",
        "composers": "A:COMPOSER",
        "tracks": "A:TRACKS",
        "playlists": "A:PLAYLISTS",
        "share": "S:",
        "sonos_playlists": "SQ:",
        "categories": "A:",
        "sonos_favorites": "FV:2",
        "radio_stations": "R:0/0",
        "radio_shows": "R:0/1",
    }

    # pylint: disable=invalid-name, protected-access
    def __init__(self, soco=None):
        """
        Args:
            soco (`SoCo`, optional): A `SoCo` instance to query for music
                library information. If `None`, or not supplied, a random
                `SoCo` instance will be used.
        """
        self.soco = soco if soco is not None else discovery.any_soco()
        self.contentDirectory = self.soco.contentDirectory

    def build_album_art_full_uri(self, url):
        """Ensure an Album Art URI is an absolute URI.

        Args:
             url (str): the album art URI.

        Returns:
            str: An absolute URI.
        """
        # Add on the full album art link, as the URI version
        # does not include the ipaddress
        if not url.startswith(("http:", "https:")):
            url = "http://" + self.soco.ip_address + ":1400" + url
        return url

    def _update_album_art_to_full_uri(self, item):
        """Update an item's Album Art URI to be an absolute URI.

        Args:
            item: The item to update the URI for
        """
        if getattr(item, "album_art_uri", False):
            item.album_art_uri = self.build_album_art_full_uri(item.album_art_uri)

    def get_artists(self, *args, **kwargs):
        """Convenience method for `get_music_library_information`
        with ``search_type='artists'``. For details of other arguments,
        see `that method
        <#soco.music_library.MusicLibrary.get_music_library_information>`_.

        """
        args = tuple(["artists"] + list(args))
        return self.get_music_library_information(*args, **kwargs)

    def get_album_artists(self, *args, **kwargs):
        """Convenience method for `get_music_library_information`
        with ``search_type='album_artists'``. For details of other arguments,
        see `that method
        <#soco.music_library.MusicLibrary.get_music_library_information>`_.

        """
        args = tuple(["album_artists"] + list(args))
        return self.get_music_library_information(*args, **kwargs)

    def get_albums(self, *args, **kwargs):
        """Convenience method for `get_music_library_information`
        with ``search_type='albums'``. For details of other arguments,
        see `that method
        <#soco.music_library.MusicLibrary.get_music_library_information>`_.

        """
        args = tuple(["albums"] + list(args))
        return self.get_music_library_information(*args, **kwargs)

    def get_genres(self, *args, **kwargs):
        """Convenience method for `get_music_library_information`
        with ``search_type='genres'``. For details of other arguments,
        see `that method
        <#soco.music_library.MusicLibrary.get_music_library_information>`_.

        """
        args = tuple(["genres"] + list(args))
        return self.get_music_library_information(*args, **kwargs)

    def get_composers(self, *args, **kwargs):
        """Convenience method for `get_music_library_information`
        with ``search_type='composers'``. For details of other arguments,
        see `that method
        <#soco.music_library.MusicLibrary.get_music_library_information>`_.

        """
        args = tuple(["composers"] + list(args))
        return self.get_music_library_information(*args, **kwargs)

    def get_tracks(self, *args, **kwargs):
        """Convenience method for `get_music_library_information`
        with ``search_type='tracks'``. For details of other arguments,
        see `that method
        <#soco.music_library.MusicLibrary.get_music_library_information>`_.

        """
        args = tuple(["tracks"] + list(args))
        return self.get_music_library_information(*args, **kwargs)

    def get_playlists(self, *args, **kwargs):
        """Convenience method for `get_music_library_information`
        with ``search_type='playlists'``. For details of other arguments,
        see `that method
        <#soco.music_library.MusicLibrary.get_music_library_information>`_.

        Note:
            The playlists that are referred to here are the playlists imported
            from the music library, they are not the Sonos playlists.

        """
        args = tuple(["playlists"] + list(args))
        return self.get_music_library_information(*args, **kwargs)

    def get_sonos_favorites(self, *args, **kwargs):
        """Convenience method for `get_music_library_information`
        with ``search_type='sonos_favorites'``. For details of other arguments,
        see `that method
        <#soco.music_library.MusicLibrary.get_music_library_information>`_.
        """
        args = tuple(["sonos_favorites"] + list(args))
        return self.get_music_library_information(*args, **kwargs)

    def get_favorite_radio_stations(self, *args, **kwargs):
        """Convenience method for `get_music_library_information`
        with ``search_type='radio_stations'``. For details of other arguments,
        see `that method
        <#soco.music_library.MusicLibrary.get_music_library_information>`_.
        """
        args = tuple(["radio_stations"] + list(args))
        return self.get_music_library_information(*args, **kwargs)

    def get_favorite_radio_shows(self, *args, **kwargs):
        """Convenience method for `get_music_library_information`
        with ``search_type='radio_stations'``. For details of other arguments,
        see `that method
        <#soco.music_library.MusicLibrary.get_music_library_information>`_.
        """
        args = tuple(["radio_shows"] + list(args))
        return self.get_music_library_information(*args, **kwargs)

        # pylint: disable=too-many-locals, too-many-arguments, too-many-branches

    def get_music_library_information(
        self,
        search_type,
        start=0,
        max_items=100,
        full_album_art_uri=False,
        search_term=None,
        subcategories=None,
        complete_result=False,
    ):
        """Retrieve music information objects from the music library.

        This method is the main method to get music information items, like
        e.g. tracks, albums etc., from the music library with. It can be used
        in a few different ways:

        The ``search_term`` argument performs a fuzzy search on that string in
        the results, so e.g calling::

            get_music_library_information('artists', search_term='Metallica')

        will perform a fuzzy search for the term 'Metallica' among all the
        artists.

        Using the ``subcategories`` argument, will jump directly into that
        subcategory of the search and return results from there. So. e.g
        knowing that among the artist is one called 'Metallica', calling::

            get_music_library_information('artists',
                                          subcategories=['Metallica'])

        will jump directly into the 'Metallica' sub category and return the
        albums associated with Metallica and::

            get_music_library_information('artists',
                                          subcategories=['Metallica', 'Black'])

        will return the tracks of the album 'Black' by the artist 'Metallica'.
        The order of sub category types is: Genres->Artists->Albums->Tracks.
        It is also possible to combine the two, to perform a fuzzy search in a
        sub category.

        The ``start``, ``max_items`` and ``complete_result`` arguments all
        have to do with paging of the results. By default the searches are
        always paged, because there is a limit to how many items we can get at
        a time. This paging is exposed to the user with the ``start`` and
        ``max_items`` arguments. So calling::

            get_music_library_information('artists', start=0, max_items=100)
            get_music_library_information('artists', start=100, max_items=100)

        will get the first and next 100 items, respectively. It is also
        possible to ask for all the elements at once::

            get_music_library_information('artists', complete_result=True)

        This will perform the paging internally and simply return all the
        items.

        Args:

            search_type (str):
                The kind of information to retrieve. Can be one of:
                ``'artists'``, ``'album_artists'``, ``'albums'``,
                ``'genres'``, ``'composers'``, ``'tracks'``, ``'share'``,
                ``'sonos_playlists'``, or ``'playlists'``, where playlists
                are the imported playlists from the music library.
            start (int, optional): starting number of returned matches
                (zero based). Default 0.
            max_items (int, optional): Maximum number of returned matches.
                Default 100.
            full_album_art_uri (bool):
                whether the album art URI should be absolute (i.e. including
                the IP address). Default `False`.
            search_term (str, optional):
                a string that will be used to perform a fuzzy search among the
                search results. If used in combination with subcategories,
                the fuzzy search will be performed in the subcategory.
            subcategories (str, optional):
                A list of strings that indicate one or more subcategories to
                dive into.
            complete_result (bool): if `True`, will disable
                paging (ignore ``start`` and ``max_items``) and return all
                results for the search.

        Warning:
            Getting e.g. all the tracks in a large collection might
            take some time.


        Returns:
             `SearchResult`: an instance of `SearchResult`.

        Note:
            * The maximum numer of results may be restricted by the unit,
              presumably due to transfer size consideration, so check the
              returned number against that requested.

            * The playlists that are returned with the ``'playlists'`` search,
              are the playlists imported from the music library, they
              are not the Sonos playlists.

        Raises:
             `SoCoException` upon errors.
        """
        search = self.SEARCH_TRANSLATION[search_type]

        # Add sub categories
        # sub categories are not allowed when searching shares
        if subcategories is not None and search_type != "share":
            for category in subcategories:
                search += "/" + url_escape_path(really_unicode(category))
        # Add fuzzy search
        if search_term is not None:
            if search_type == "share":
                # Don't insert ":" and don't escape "/" (so can't use url_escape_path)
                search += quote_url(really_unicode(search_term).encode("utf-8"))
            else:
                search += ":" + url_escape_path(really_unicode(search_term))

        item_list = []
        metadata = {"total_matches": 100000}
        while len(item_list) < metadata["total_matches"]:
            # Change start and max for complete searches
            if complete_result:
                start, max_items = len(item_list), 100000

            # Try and get this batch of results
            try:
                response, metadata = self._music_lib_search(search, start, max_items)
            except SoCoUPnPException as exception:
                # 'No such object' UPnP errors
                if exception.error_code == "701":
                    return SearchResult([], search_type, 0, 0, None)
                else:
                    raise exception

            # Parse the results
            items = from_didl_string(response["Result"])
            for item in items:
                # Check if the album art URI should be fully qualified
                if full_album_art_uri:
                    self._update_album_art_to_full_uri(item)
                # Append the item to the list
                item_list.append(item)

            # If we are not after the complete results, the stop after 1
            # iteration
            if not complete_result:
                break

        metadata["search_type"] = search_type
        if complete_result:
            metadata["number_returned"] = len(item_list)

        # pylint: disable=star-args
        return SearchResult(item_list, **metadata)

    def browse(
        self,
        ml_item=None,
        start=0,
        max_items=100,
        full_album_art_uri=False,
        search_term=None,
        subcategories=None,
    ):
        """Browse (get sub-elements from) a music library item.

        Args:
            ml_item (`DidlItem`): the item to browse, if left out or
                `None`, items at the root level will be searched.
            start (int): the starting index of the results.
            max_items (int): the maximum number of items to return.
            full_album_art_uri (bool): whether the album art URI should be
                fully qualified with the relevant IP address.
            search_term (str): A string that will be used to perform a fuzzy
                search among the search results. If used in combination with
                subcategories, the fuzzy search will be performed on the
                subcategory. Note: Searching will not work if ``ml_item`` is
                `None`.
            subcategories (list): A list of strings that indicate one or more
                subcategories to descend into. Note: Providing sub categories
                will not work if ``ml_item`` is `None`.

        Returns:
            A `SearchResult` instance.

        Raises:
            AttributeError: if ``ml_item`` has no ``item_id`` attribute.
            SoCoUPnPException: with ``error_code='701'`` if the item cannot be
                browsed.
        """
        if ml_item is None:
            search = "A:"
        else:
            search = ml_item.item_id

        # Add sub categories
        if subcategories is not None:
            for category in subcategories:
                search += "/" + url_escape_path(really_unicode(category))
        # Add fuzzy search
        if search_term is not None:
            search += ":" + url_escape_path(really_unicode(search_term))

        try:
            response, metadata = self._music_lib_search(search, start, max_items)
        except SoCoUPnPException as exception:
            # 'No such object' UPnP errors
            if exception.error_code == "701":
                return SearchResult([], "browse", 0, 0, None)
            else:
                raise exception
        metadata["search_type"] = "browse"

        # Parse the results
        containers = from_didl_string(response["Result"])
        item_list = []
        for container in containers:
            # Check if the album art URI should be fully qualified
            if full_album_art_uri:
                self._update_album_art_to_full_uri(container)
            item_list.append(container)

        # pylint: disable=star-args
        return SearchResult(item_list, **metadata)

    def browse_by_idstring(
        self, search_type, idstring, start=0, max_items=100, full_album_art_uri=False
    ):
        """Browse (get sub-elements from) a given music library item,
        specified by a string.

        Args:
            search_type (str): The kind of information to retrieve. Can be
                one of: ``'artists'``, ``'album_artists'``, ``'albums'``,
                ``'genres'``, ``'composers'``, ``'tracks'``, ``'share'``,
                ``'sonos_playlists'``, and ``'playlists'``, where
                playlists are the imported file based playlists from the
                music library.
            idstring (str): a term to search for.
            start (int): starting number of returned matches. Default 0.
            max_items (int): Maximum number of returned matches. Default 100.
            full_album_art_uri (bool): whether the album art URI should be
                absolute (i.e. including the IP address). Default `False`.

        Returns:
            `SearchResult`: a `SearchResult` instance.

        Note:
            The maximum numer of results may be restricted by the unit,
            presumably due to transfer size consideration, so check the
            returned number against that requested.
        """
        search = self.SEARCH_TRANSLATION[search_type]

        # Check if the string ID already has the type, if so we do not want to
        # add one also Imported playlist have a full path to them, so they do
        # not require the A:PLAYLISTS part first
        if idstring.startswith(search) or (search_type == "playlists"):
            search = ""

        search_item_id = search + idstring
        search_uri = "#" + search_item_id
        # Not sure about the res protocol. But this seems to work
        res = [DidlResource(uri=search_uri, protocol_info="x-rincon-playlist:*:*:*")]
        search_item = DidlObject(
            resources=res, title="", parent_id="", item_id=search_item_id
        )

        # Call the base version
        return self.browse(search_item, start, max_items, full_album_art_uri)

    def _music_lib_search(self, search, start, max_items):
        """Perform a music library search and extract search numbers.

        You can get an overview of all the relevant search prefixes (like
        'A:') and their meaning with the request:

        .. code ::

         response = device.contentDirectory.Browse([
             ('ObjectID', '0'),
             ('BrowseFlag', 'BrowseDirectChildren'),
             ('Filter', '*'),
             ('StartingIndex', 0),
             ('RequestedCount', 100),
             ('SortCriteria', '')
         ])

        Args:
            search (str): The ID to search.
            start (int): The index of the forst item to return.
            max_items (int): The maximum number of items to return.

        Returns:
            tuple: (response, metadata) where response is the returned metadata
                and metadata is a dict with the 'number_returned',
                'total_matches' and 'update_id' integers
        """
        response = self.contentDirectory.Browse(
            [
                ("ObjectID", search),
                ("BrowseFlag", "BrowseDirectChildren"),
                ("Filter", "*"),
                ("StartingIndex", start),
                ("RequestedCount", max_items),
                ("SortCriteria", ""),
            ]
        )

        # Get result information
        metadata = {}
        for tag in ["NumberReturned", "TotalMatches", "UpdateID"]:
            metadata[camel_to_underscore(tag)] = int(response[tag])
        return response, metadata

    @property
    def library_updating(self):
        """bool: whether the music library is in the process of being updated."""
        result = self.contentDirectory.GetShareIndexInProgress()
        return result["IsIndexing"] != "0"

    def start_library_update(self, album_artist_display_option=""):
        """Start an update of the music library.

        Args:
            album_artist_display_option (str): a value for the album artist
                compilation setting (see `album_artist_display_option`).
        """
        return self.contentDirectory.RefreshShareIndex(
            [
                ("AlbumArtistDisplayOption", album_artist_display_option),
            ]
        )

    def search_track(self, artist, album=None, track=None, full_album_art_uri=False):
        """Search for an artist, an artist's albums, or specific track.

        Args:
            artist (str): an artist's name.
            album (str, optional): an album name. Default `None`.
            track (str, optional): a track name. Default `None`.
            full_album_art_uri (bool): whether the album art URI should be
                absolute (i.e. including the IP address). Default `False`.

        Returns:
            A `SearchResult` instance.
        """
        subcategories = [artist]
        subcategories.append(album or "")

        # Perform the search
        result = self.get_album_artists(
            full_album_art_uri=full_album_art_uri,
            subcategories=subcategories,
            search_term=track,
            complete_result=True,
        )
        result._metadata["search_type"] = "search_track"
        return result

    def get_albums_for_artist(self, artist, full_album_art_uri=False):
        """Get an artist's albums.

        Args:
            artist (str): an artist's name.
            full_album_art_uri: whether the album art URI should be
                absolute (i.e. including the IP address). Default `False`.

        Returns:
            A `SearchResult` instance.
        """
        subcategories = [artist]
        result = self.get_album_artists(
            full_album_art_uri=full_album_art_uri,
            subcategories=subcategories,
            complete_result=True,
        )

        reduced = [item for item in result if item.__class__ == DidlMusicAlbum]
        # It is necessary to update the list of items in two places, due to
        # a bug in SearchResult
        result[:] = reduced
        result._metadata.update(
            {
                "item_list": reduced,
                "search_type": "albums_for_artist",
                "number_returned": len(reduced),
                "total_matches": len(reduced),
            }
        )
        return result

    def get_tracks_for_album(self, artist, album, full_album_art_uri=False):
        """Get the tracks of an artist's album.

        Args:
            artist (str): an artist's name.
            album (str): an album name.
            full_album_art_uri: whether the album art URI should be
                absolute (i.e. including the IP address). Default `False`.

        Returns:
            A `SearchResult` instance.
        """
        subcategories = [artist, album]
        result = self.get_album_artists(
            full_album_art_uri=full_album_art_uri,
            subcategories=subcategories,
            complete_result=True,
        )
        result._metadata["search_type"] = "tracks_for_album"
        return result

    @property
    def album_artist_display_option(self):
        """str: The current value of the album artist compilation setting.

        Possible values are:

        * ``'WMP'`` - use Album Artists
        * ``'ITUNES'`` - use iTunes® Compilations
        * ``'NONE'`` - do not group compilations

        See Also:
            The Sonos `FAQ <https://sonos.custhelp.com
            /app/answers/detail/a_id/3056/kw/artist%20compilation>`_ on
            compilation albums.

        To change the current setting, call `start_library_update` and
        pass the new setting.
        """
        result = self.contentDirectory.GetAlbumArtistDisplayOption()
        return result["AlbumArtistDisplayOption"]

    def list_library_shares(self):
        """Return a list of the music library shares.

        Returns:
            list: The music library shares, which are strings of the form
            ``'//hostname_or_IP/share_path'``.
        """
        response = self.contentDirectory.Browse(
            [
                ("ObjectID", "S:"),
                ("BrowseFlag", "BrowseDirectChildren"),
                ("Filter", "*"),
                ("StartingIndex", "0"),
                ("RequestedCount", "100"),
                ("SortCriteria", ""),
            ]
        )
        shares = []
        matches = response["TotalMatches"]
        # Zero matches
        if matches == "0":
            return shares
        xml_dict = xmltodict.parse(response["Result"])
        unpacked = xml_dict["DIDL-Lite"]["container"]
        # One match
        if matches == "1":
            shares.append(unpacked["dc:title"])
            return shares
        # Otherwise it's multiple matches
        for share in unpacked:
            shares.append(share["dc:title"])
        return shares

    def delete_library_share(self, share_name):
        """Delete a music library share.

        Args:
            share_name (str): the name of the share to be deleted, which
                should be of the form ``'//hostname_or_IP/share_path'``.

        :raises: `SoCoUPnPException`
        """
        # share_name must be prefixed with 'S:'
        self.contentDirectory.DestroyObject([("ObjectID", "S:" + share_name)])



"""Tests for the alarms module."""


from pysonos.alarms import is_valid_recurrence


def test_recurrence():
    for recur in ("DAILY", "WEEKDAYS", "WEEKENDS", "ONCE"):
        assert is_valid_recurrence(recur)

    assert is_valid_recurrence("ON_1")
    assert is_valid_recurrence("ON_123412")
    assert not is_valid_recurrence("on_1")
    assert not is_valid_recurrence("ON_123456789")
    assert not is_valid_recurrence("ON_")
    assert not is_valid_recurrence(" ON_1")
